/*
******************************************************************************
*
*      File             : TFO.cpp       (was: tfo_main.c)
*
*      Version          : 5.0.0  
*      Date             : 22 February 2002 
*      Author           : Clemens Suerbaum
*
*      Changes to former versions can be found by looking for the string of the
*      current version("5.0.0")
*      Changes in pure formating or for better readability of the code were not 
*      marked.
*
*
*      Former Versions  : 2.0.0  
*      Date             : 27 February 2001
*      Author           : Jochen Junkawitsch
*
*      Description      : source file for TFO decision algorithm
*
******************************************************************************
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "tfo_decision.h"
#include "extensionsForAMRWB.h"           // 5.0.0
#include <iostream.h>                     // 5.0.0




// ------------------------------------------------------------------------
//
// Function  s2cs
// Purpose   convert a string (example "-X--XX-X") to a CodecSet 
// 
// this function converts a string into a value representing a codec set
// the string must have a length of 8
// an "x" at the n-th position adds the (n-1)th power of 2 to the codec 
// set value
// ------------------------------------------------------------------------

CodecSet s2cs(char* s)
{
  int k;
  CodecSet m, cs = 0;
  
  if (strlen(s) != 8)
	{fprintf(stderr,"ERROR in input: l/dACS/SCS not 8 characters long\n");exit(0);} //4.3.0 error text
  
  for (k=0, m=mode_122; k<8; k++, m >>= 1)
    if ( (s[k] == 'X') || (s[k] == 'x') )   // 4.3.0 also small x allowed
      cs |= m;
  
  return cs;
}

// ------------------------------------------------------------------------
//
// Function  wb_s2cs
// Purpose   convert a string (example "-X--XX-X-") to a WB_CodecSet 
// 
// New in 5.0.0 (extension for AMR-WB)
// 
// ------------------------------------------------------------------------

WB_CodecSet wb_s2cs(char* s)
{
  if (strlen(s) != 9)
	{fprintf(stderr,"ERROR in input: l/dACS/SCS not 9 characters long\n");exit(0);}  
  
  int k = 0;;
  WB_CodecSet cs = 0;
  WB_CodecSet m = wb_mode_2385;    // m represents the value of the mode to be checked
                          // start with highest mode
  
  for (k=0; k<9; k++) 
	{
	  if ( (s[k] == 'x') || (s[k] == 'X') )                       
		cs = cs + m;    // adds mode to the codec set
		
	  m = m/2;          // moves m to value for next LOWER mode in the loop
	}
  
  return cs;
}


// ------------------------------------------------------------------------
//
// Function  getInputData
// Purpose   read input Data from stdin 
// 
// ------------------------------------------------------------------------

int getInputData(Data* data)
{
	char buffer[255];

	char lscs[255], dscs[255];
	char lacs[255], dacs[255];
	int  lmacs, dmacs;
	char lType[255], dType[255];
	char lOM, dOM;

	// initialisation:
	data->cacs = data->oacs = data->cscs = 0;
	//5.0.0 from here on new
	data->wb_cacs = data->wb_oacs = data->wb_cscs = 0;
	data->wb_lacs = data->wb_lscs =0;    
	data->lacs = data->lscs = 0;

	data->wb_dacs = data->wb_dscs =0; 
	data->dacs = data->dscs = 0;

	data->lType = data->dType = dummy;
	data->lOM=data->dOM=0;
    data->lmacs=data->dmacs=data->cmacs=0;

	data->wb_default_acs = 0;
	data->evaluation_wb_oacs_step1 = 0;
	data->evaluation_wb_oacs_step2 = 0;
	//5.0.0 end of new insert

	//-----------------------------------------------------------5.0.0
	// Output text for input
	//-----------------------------------------------------------
	// 4.3.0 for easier input

	printf("----------------------------------------------------------------------\n\n"); 
	printf("TFO 3GPP TS28.062 V5.0.0 decision algorithm reference implementation \n\n"); 
	printf("This program checks if TFO is possible and, if on AMR, on which CACS. \n"); 
	printf("Expected input: \n"); 
	printf("lSCS lACS lMACS lType lOM dSCS dACS dMACS dType dOM   (l: local, d: distant)\n");
	printf("For all codec sets (CS) highest mode comes first.\n");
	printf("Every CS must consist of 8 characters for AMR-NB, 9 for AMR-WB.\n");
	printf("OMs are y(es) or n(o). MACSs = 1..8 for AMR-NB and AMR-WB.\n");
	printf("x or X: mode present, - (or any other character): not present\n"); 
	printf("Examples:\n");
	printf("-------- -------- 0 GSM_EFR y -------- -------- 0 GSM_EFR n \n"); 
	printf("xxxxxxxx x-x-x-x- 4 FR_AMR y --xxxxxx ---x-x-x 3 HR_AMR n \n"); 
	printf("xxxxxxxx -x-x-x-x 4 UMTS_AMR y xxxxxxxx --x-x-x- 3 UMTS_AMR2 n \n"); 
	// 5.0.0 adds:
	printf("xxxxxxxxx x-x-x-x-x 5 UMTS_AMR_WB y --xxxxxxx ----x-x-x 4 FR_AMR_WB y \n"); 
	printf("xxxxxxxxx x-x-x-x-x 5 UMTS_AMR_WB y xxxxxxxxx -x-x-x-x- 4 UMTS_AMR_WB y \n"); 
	// end insert 5.0.0
	printf("Your input, please: [to exit push the Return button two times,\n");
	printf("                     but close your eyes before doing so ;-) ]\n"); 
	printf(" \n");


	//-----------------------------------------------------------5.0.0
	// Read input
	//-----------------------------------------------------------
	if (!fgets(buffer, 255, stdin))
	  return 0;

	printf("\n"); 
 	printf("Reaction on your input:\n"); 
    //- - - - - - - - - - - - - - - - - - - - - - - - - - -5.0.0
	// Check number of input parameters
	//- - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (sscanf(buffer, "%s %s %d %s %c %s %s %d %s %c",
			 lscs, lacs, &lmacs, lType, &lOM,                   
			 dscs, dacs, &dmacs, dType, &dOM) != 10)
	 {fprintf(stderr,"ERROR in input: not exactly 10 parameters\n");exit(0);} // 5.0.0 unique error text

	//-----------------------------------------------------------5.0.0
	// Convert codec type (local)
	//-----------------------------------------------------------
	if (!strcmp(lType, "FR_AMR"))			// strcmp=0 means: identical
		data->lType = FR_AMR;
	else if (!strcmp(lType, "HR_AMR"))
		data->lType = HR_AMR;
	else if (!strcmp(lType, "UMTS_AMR"))
		data->lType = UMTS_AMR;
	else if (!strcmp(lType, "UMTS_AMR_2"))
		data->lType = UMTS_AMR_2;
	else if (!strcmp(lType, "GSM_EFR"))
		data->lType = GSM_EFR;
	else if (!strcmp(lType, "GSM_FR"))
		data->lType = GSM_FR;
	else if (!strcmp(lType, "GSM_HR"))
		data->lType = GSM_HR;
						 // 5.0.0 begin: add AMR-WB codec types
	else if (!strcmp(lType, "OHR_AMR"))
		data->lType = OHR_AMR;
	else if (!strcmp(lType, "FR_AMR_WB"))
		data->lType = FR_AMR_WB;
	else if (!strcmp(lType, "UMTS_AMR_WB"))
		data->lType = UMTS_AMR_WB;
	else if (!strcmp(lType, "OFR_AMR_WB"))
		data->lType = OFR_AMR_WB;
	else if (!strcmp(lType, "OHR_AMR_WB"))
		data->lType = OHR_AMR_WB;
						 // 5.0.0 end of: add AMR-WB codec types
	else
		{
		fprintf(stderr,"ERROR in input: undefined local codec tpye\n"); // 5.0.0 unique error text
		exit(0);
		}  
	
    //- - - - - - - - - - - - - - - - - - - - - - - - - - -5.0.0
	// Check correct range for MACS (local) and read lscs/lacs
	//- - - - - - - - - - - - - - - - - - - - - - - - - - -
	// 5.0.0  this part works only for AMR narrowband
	if (
		(data->lType==HR_AMR    ) ||		
		(data->lType==FR_AMR    ) || 
		(data->lType==UMTS_AMR  ) || 
		(data->lType==UMTS_AMR_2) || 
		(data->lType==OHR_AMR)  
		)
		{
		// 5.0.0   this block moved here from above codec type check
		data->lscs  = s2cs(lscs);
		data->lacs  = s2cs(lacs);

		if (lmacs > 0 && lmacs <= 8)
			data->lmacs = lmacs;
		else
			{
			fprintf(stderr,"ERROR in input: lmacs outside allowed range 1..8\n"); // 5.0.0 unique error text
			exit(0);}    
			}	 

	// 5.0.0  begin: extensions for AMR-WB
	if (
		(data->lType == FR_AMR_WB    )  ||		
		(data->lType == UMTS_AMR_WB  )  ||
		(data->lType == OHR_AMR_WB   )  ||
		(data->lType == OFR_AMR_WB   )
		)
		{   
		data->lscs = data->lacs = 0;               // AMR codec sets 

		data->wb_lscs  = wb_s2cs(lscs);
		
		// check: only lower seven modes in scs for FR_AMR_WB
		if (
			(data->lType == FR_AMR_WB) && (data->wb_lscs > 255)
			)
			{
			fprintf(stderr,"ERROR in input: highest two modes not allowed for FR_AMR_WB (lSCS)\n"); 
			exit(0);   
			}

		data->wb_lacs  = wb_s2cs(lacs);

		
		// check: only lower seven modes in ACS for FR_AMR_WB
		if (
			(data->lType == FR_AMR_WB) && (data->wb_lacs > 255)
			)
			{
			fprintf(stderr,"ERROR in input: highest two modes not allowed for FR_AMR_WB (lACS)\n"); 
			exit(0);   
			}

			
		// check: range of lmacs
		if ( (lmacs > 0) && (lmacs <= max_wb_macs) )
			data->lmacs = lmacs;
		else
			{
			fprintf(stderr,"ERROR in input: lmacs outside allowed range (1..max_wb_macs=8) for AMR-WB \n");
			exit(0);
			} 

		
		// check: consistency of lacs and lmacs
		if ( noOfWBmodes(data->wb_lacs) > data->lmacs )
			{
			fprintf(stderr,"ERROR in input: more modes in ACS than MACS (local)\n"); 
			exit(0);   
			}
		
		
		// check: MaMiMo (Mandatory Minimum Set of Modes) for AMR-WB
		if ( 
			!wb_mode_isInWBcodecSet(wb_mode_0885,data->wb_lscs) ||
			!wb_mode_isInWBcodecSet(wb_mode_1265,data->wb_lscs) 
			)
			{
			fprintf(stderr,"MaMiMo (8,85 and 12,65) for AMR-WB not supported (local)\n"); 
			exit(0);   
			}

		} 
	// 5.0.0  end: extensions for AMR-WB

    //- - - - - - - - - - - - - - - - - - - - - - - - - - -5.0.0
	// Check and convert input for OM (local)
	//- - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (lOM == 'y')
		data->lOM = 1;
	else if (lOM == 'n')
		data->lOM = 0;
	else
		{
		fprintf(stderr,"ERROR in input: lOM not 'y' or 'n'\n"); // 5.0.0 unique error text
		exit(0);
		} 


	//-----------------------------------------------------------5.0.0
	// Convert codec type (distant)
	//-----------------------------------------------------------
	if (!strcmp(dType, "FR_AMR"))			// strcmp=0 means: identical
		data->dType = FR_AMR;
	else if (!strcmp(dType, "HR_AMR"))
		data->dType = HR_AMR;
	else if (!strcmp(dType, "UMTS_AMR"))
		data->dType = UMTS_AMR;
	else if (!strcmp(dType, "UMTS_AMR_2"))
		data->dType = UMTS_AMR_2;
	else if (!strcmp(dType, "GSM_EFR"))
		data->dType = GSM_EFR;
	else if (!strcmp(dType, "GSM_FR"))
		data->dType = GSM_FR;
	else if (!strcmp(dType, "GSM_HR"))
		data->dType = GSM_HR;
						// 5.0.0 begin: add AMR-WB codec types
	else if (!strcmp(dType, "OHR_AMR"))
		data->dType = OHR_AMR;
	else if (!strcmp(dType, "FR_AMR_WB"))
		data->dType = FR_AMR_WB;
	else if (!strcmp(dType, "UMTS_AMR_WB"))
		data->dType = UMTS_AMR_WB;
	else if (!strcmp(dType, "OFR_AMR_WB"))
		data->dType = OFR_AMR_WB;
	else if (!strcmp(dType, "OHR_AMR_WB"))
		data->dType = OHR_AMR_WB;
					   // 5.0.0 end of: add AMR-WB codec types
	else
		{
		fprintf(stderr,"ERROR in input: undefined distant codec tpye\n"); // 5.0.0 unique error text
		exit(0);
		}  

	//- - - - - - - - - - - - - - - - - - - - - - - - - - -5.0.0
	// Check correct range for MACS (distant)
	//- - - - - - - - - - - - - - - - - - - - - - - - - - -
	// 5.0.0  this part works only for AMR narrowband
	if (
		(data->dType == HR_AMR     ) ||           
		(data->dType == FR_AMR     ) || 
		(data->dType == UMTS_AMR   ) || 
		(data->dType == UMTS_AMR_2 ) || 
		(data->dType == OHR_AMR)  
		)
		{
		// 5.0.0   this block moved here from above codec type check 
		data->dacs  = s2cs(dacs);
		data->dscs  = s2cs(dscs);  

		if ( (dmacs > 0) && (dmacs <= 8) )
			data->dmacs = dmacs;
		else
			{
			fprintf(stderr,"ERROR in input: dmacs outside allowed range (1..8)\n"); // 5.0.0 unique error text
			exit(0);
			}  
		}

	// 5.0.0  begin: extensions for AMR-WB
	if (
		(data->dType == FR_AMR_WB    )  ||         
		(data->dType == UMTS_AMR_WB  )  ||
		(data->dType == OHR_AMR_WB   )  ||
		(data->dType == OFR_AMR_WB   )
		)
		{   
		data->dscs = data->dacs = 0;               // AMR codec sets 

		data->wb_dscs  = wb_s2cs(dscs);

				
		// check: only lower seven modes in SCS for FR_AMR_WB
		if (
			(data->dType == FR_AMR_WB) && (data->wb_dscs > 255)
			)
			{
			fprintf(stderr,"ERROR in input: highest two modes not allowed for FR_AMR_WB (dSCS)\n"); 
			exit(0);   
			}

		data->wb_dacs  = wb_s2cs(dacs);
		
		
		// check: only lower seven modes in ACS for FR_AMR_WB
		if (
			(data->dType == FR_AMR_WB) && (data->wb_dacs > 255)
			)
			{
			fprintf(stderr,"ERROR in input: highest two modes not allowed for FR_AMR_WB (dACS)\n"); 
			exit(0);   
			}

		// check: range of dmacs
			if (dmacs > 0 && dmacs <= max_wb_macs)
			data->dmacs = dmacs;
		else
			{
			fprintf(stderr,"ERROR in input: dmacs outside allowed range (1..max_wb_macs=8) for AMR-WB\n");
			exit(0);
			} 

		
		// check: consistency of dacs and dmacs
		if ( noOfWBmodes(data->wb_dacs) > data->dmacs )
			{
			fprintf(stderr,"ERROR in input: more modes in ACS than MACS (distant)\n"); 
			exit(0);   
			}
			
		// check: MaMiMo (Mandatory Minimum Setof Modes) for AMR-WB
		if ( 
			!wb_mode_isInWBcodecSet(wb_mode_1265,data->wb_dscs) ||
			!wb_mode_isInWBcodecSet(wb_mode_1265,data->wb_dscs) 
			)
			{
			fprintf(stderr,"MaMiMo (8,85 and 12,65) for AMR-WB not supported (distant)\n"); 
			exit(0);   
			}
			
		} 
	// 5.0.0  end: extensions for AMR-WB
 

    //- - - - - - - - - - - - - - - - - - - - - - - - - - -5.0.0
	// Check and convert input for OM (distant)
	//- - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (dOM == 'y')
		data->dOM = 1;
	else if (dOM == 'n')
		data->dOM = 0;
	else
		{
		fprintf(stderr,"ERROR in input: dOM not 'y' or 'n'\n"); // 5.0.0 unique error text
		exit(0);
		} 


    //--------------------------------------------------------------5.0.0
	// set AMR parameters to zero for non-AMR codec types (local)
	//--------------------------------------------------------------
	if (
		(data->lType == GSM_EFR)  ||     // 5.0.0 changed to explicit list
		(data->lType == GSM_FR )  || 
		(data->lType == GSM_HR )   
		)
		{
		/*data->lOM =*/ data->lscs = data->lacs =  data->lmacs = 0;   //4.3.0 lOM untouched
		data->wb_lscs = data->wb_lacs = 0;  // 5.0.0
		}

	//--------------------------------------------------------------5.0.0
	// set AMR parameters to zero for non-AMR codec types (distant)
	//--------------------------------------------------------------
	if (
		(data->dType == GSM_EFR)  ||     // 5.0.0 changed to explicit list
		(data->dType == GSM_FR )  || 
		(data->dType == GSM_HR )   
		)
		{
		/*data->dOM =*/ data->dscs = data->dacs = data->dmacs = 0;   // 4.3.0 dOM untouched
		data->wb_dscs = data->wb_dacs = 0;  // 5.0.0
		}


	//-----------------------------------------------------------
	// check that local ACS is part of local SCS
	//-----------------------------------------------------------

	// 5.0.0  this part works only for AMR narrowband
	if (
		(lType=="HR_AMR"    )  || 
		(lType=="FR_AMR"    )  || 
		(lType=="UMTS_AMR"  )  || 
		(lType=="UMTS_AMR_2")  || 
		(lType=="OHR_AMR")  
		)
		{
		if ( (data->lacs & data->lscs) != data->lacs  )
			{fprintf(stderr,"ERROR in input: lACS not in lSCS\n");exit(0);}

		if ( (data->dacs & data->dscs) != data->dacs    )      // 5.0.0 separate cases
			{fprintf(stderr,"ERROR in input: dACS not in dSCS\n");exit(0);}
		
		data->wb_lscs = data->wb_lacs = 0;  // 5.0.0
		}

    // 5.0.0  begin: extensions for AMR-WB
	else if (lType=="FR_AMR_WB"  || 
		     lType=="UMTS_AMR_WB"||
			 lType=="OFR_AMR_WB" ||
			 lType=="OHR_AMR_WB"
		)
		{   
		if (
			((wb_mode_isInWBcodecSet(wb_mode_2385,data->lacs)==1)&&
			 (wb_mode_isInWBcodecSet(wb_mode_2385,data->lscs)==0)  ) ||
            ((wb_mode_isInWBcodecSet(wb_mode_2305,data->lacs)==1)&&
			 (wb_mode_isInWBcodecSet(wb_mode_2305,data->lscs)==0)  ) ||
			((wb_mode_isInWBcodecSet(wb_mode_1985,data->lacs)==1)&&
			 (wb_mode_isInWBcodecSet(wb_mode_1985,data->lscs)==0)  ) ||
			((wb_mode_isInWBcodecSet(wb_mode_1825,data->lacs)==1)&&
			 (wb_mode_isInWBcodecSet(wb_mode_1825,data->lscs)==0)  ) ||
			((wb_mode_isInWBcodecSet(wb_mode_1585,data->lacs)==1)&&
			 (wb_mode_isInWBcodecSet(wb_mode_1585,data->lscs)==0)  ) ||
			((wb_mode_isInWBcodecSet(wb_mode_1425,data->lacs)==1)&&
			 (wb_mode_isInWBcodecSet(wb_mode_1425,data->lscs)==0)  ) ||
			((wb_mode_isInWBcodecSet(wb_mode_1265,data->lacs)==1)&&
			 (wb_mode_isInWBcodecSet(wb_mode_1265,data->lscs)==0)  ) ||
			((wb_mode_isInWBcodecSet(wb_mode_0885,data->lacs)==1)&&
			 (wb_mode_isInWBcodecSet(wb_mode_0885,data->lscs)==0)  ) ||
			((wb_mode_isInWBcodecSet(wb_mode_0660,data->lacs)==1)&&
			 (wb_mode_isInWBcodecSet(wb_mode_0660,data->lscs)==0)  ) 
		   )
			{
			fprintf(stderr,"ERROR in input: lACS not in lSCS (WB)\n");
			exit(0);
			}
		
		if (
			((wb_mode_isInWBcodecSet(256,data->dacs)==1)&&
			 (wb_mode_isInWBcodecSet(256,data->dscs)==0)  ) ||
            ((wb_mode_isInWBcodecSet(128,data->dacs)==1)&&
			 (wb_mode_isInWBcodecSet(128,data->dscs)==0)  ) ||
			((wb_mode_isInWBcodecSet( 64,data->dacs)==1)&&
			 (wb_mode_isInWBcodecSet( 64,data->dscs)==0)  ) ||
			((wb_mode_isInWBcodecSet( 32,data->dacs)==1)&&
			 (wb_mode_isInWBcodecSet( 32,data->dscs)==0)  ) ||
			((wb_mode_isInWBcodecSet( 16,data->dacs)==1)&&
			 (wb_mode_isInWBcodecSet( 16,data->dscs)==0)  ) ||
			((wb_mode_isInWBcodecSet(  8,data->dacs)==1)&&
			 (wb_mode_isInWBcodecSet(  8,data->dscs)==0)  ) ||
			((wb_mode_isInWBcodecSet(  4,data->dacs)==1)&&
			 (wb_mode_isInWBcodecSet(  4,data->dscs)==0)  ) ||
			((wb_mode_isInWBcodecSet(  2,data->dacs)==1)&&
			 (wb_mode_isInWBcodecSet(  2,data->dscs)==0)  ) ||
			((wb_mode_isInWBcodecSet(  1,data->dacs)==1)&&
			 (wb_mode_isInWBcodecSet(  1,data->dscs)==0)  ) 
		   )
			{
			fprintf(stderr,"ERROR in input: dACS not in dSCS (WB)\n");
			exit(0);
			}
		}
	// 5.0.0  end: extensions for AMR-WB


	//------------------------------------------------------------------------
	// check number of modes in ACS does not exceed MACS (local and distant)
	//------------------------------------------------------------------------

    // 5.0.0  this part works only for AMR narrowband
	if (
		(lType=="HR_AMR"    ) || 
	    (lType=="FR_AMR"    ) || 
	    (lType=="UMTS_AMR"  ) || 
	    (lType=="UMTS_AMR_2") || 
		(lType=="OHR_AMR")   
		)
		{
		if ( data->lmacs < nModes(data->lacs) )
			{
			fprintf(stderr,"ERROR in input: lMACS < number of modes in lACS\n");
			exit(0);
			}

		if ( data->dmacs < nModes(data->dacs) )        // 5.0.0 separate cases
			{
			fprintf(stderr,"ERROR in input: dMACS < number of modes in dACS\n");
			exit(0);
			}	
		}

	// 5.0.0  begin: extensions for AMR-WB
	else if (lType=="FR_AMR_WB"  || 
		     lType=="UMTS_AMR_WB"||
			 lType=="OFR_AMR_WB" ||
			 lType=="OHR_AMR_WB"  
		)
		{   
		if ( data->lmacs < noOfWBmodes(data->lacs) )
			{
			fprintf(stderr,"ERROR in input: lMACS < number of modes in lACS (WB)\n");
			exit(0);
			}

		if ( data->dmacs < noOfWBmodes(data->dacs) ) 
			{
			fprintf(stderr,"ERROR in input: dMACS < number of modes in dACS (WB)\n");
			exit(0);
			}	
		}
	// 5.0.0  end: extensions for AMR-WB


	//---------------------------------------------------------------------------------
	// ensure that upper two modes of AMR not in ACS/SCS for HR_AMR (local and distant)
	//---------------------------------------------------------------------------------
    // changed from internal change to input error messag3e
	// 2.0.0 was: if (data->lType == HR_AMR)
	//             	{data->lscs &= 0x3f; data->lacs &= 0x3f;}  

	// 2.0.0 was: if (data->dType == HR_AMR)
	//          	{data->dscs &= 0x3f; data->dacs &= 0x3f;}
	if (
		(data->lType == HR_AMR) && 
		( (data->lacs > 127) || (data->lscs > 127) )
		)
		{
		fprintf(stderr,"ERROR in input: highest two modes not allowed for HR_AMR (local)\n"); 
		exit(0);   
		}
	
	if (
		(data->dType == HR_AMR) && 
		( (data->dacs > 127) || (data->dscs > 127) )
		)
		{
		fprintf(stderr,"ERROR in input: highest two modes not allowed for HR_AMR (distant)\n"); 
		exit(0);   
		}

    //-----------------------------------------------------------
	

	return 1;  
}



// ------------------------------------------------------------------------
//
// Function  printData
// Purpose   print Data to stdout 
// 
// ------------------------------------------------------------------------

int printData(Data* data)
{
	int k;
	CodecSet m;
	int wb_m;    // 5.0.0 added for AMR-WB
	
	// 5.0.0  this part works only for AMR narrowband
	/* Two compatible AMR codec types? */   // 4.3.0 condition for mode output introduced
	if (
		compatible_AMRNB_codec_types(data->lType,data->dType) //5.0.0 function introduced            
		)
		{
		//---------------------------------------------------------------
		// output header
		//---------------------------------------------------------------

		printf("\n\n------------------------------------------------------------------");
		printf("\n\n\t%s\t\t\t\t\t%s\n", codecName(data->lType), codecName(data->dType));
		printf("\tMACS = %d\t\t\t\tMACS = %d\n", data->lmacs, data->dmacs);  // 5.0.0 tab o.k.??
		printf("\tOM   = %s\t\t\t\tOM   = %s\n", data->lOM ? "yes" : "no", data->dOM ? "yes" : "no");
		printf("\n\tSCS\tACS\tIACS\tOACS\tCSCS\tACS\tSCS\n");

		//---------------------------------------------------------------
		// output of real data
		//---------------------------------------------------------------
		m = 128;      // start with highest mode, 5.0.0 initialisation was missing in 2.0.0

		for (k=0; m; m>>=1)   
				{
				printf( "%s\t%c\t%c\t%c\t%c\t%c\t%c\t%c\n",
					 modeName((CodecMode)m),   
					 data->lscs & m ? 'X' : '-',
					 data->lacs & m ? 'X' : '-',
					 data->cacs & m ? 'X' : '-',
					 data->oacs & m ? 'X' : '-',
					 data->cscs & m ? 'X' : '-',
					 data->dacs & m ? 'X' : '-',
					 data->dscs & m ? 'X' : '-');
				}
		}  


	// 5.0.0  begin: output extensions for AMR-WB
	else if (
		compatible_AMR_WB_codec_types(data->lType, data->dType)
		)		
		{
		//---------------------------------------------------------------
		// output header
		//---------------------------------------------------------------

		printf("\n\n------------------------------------------------------------------");
		printf("\n\n\t%s\t\t\t\t\t%s\n", codecName(data->lType), codecName(data->dType));
		printf("\tMACS = %d\t\t\t\t\tMACS = %d\n", data->lmacs, data->dmacs);  // 5.0.0 tab o.k.??
		printf("\tOM   = %s\t\t\t\t\tOM   = %s\n", data->lOM ? "yes" : "no", data->dOM ? "yes" : "no");
		printf("\n\tSCS\tACS\tIACS\tOACS\tDefACS\tCSCS\tACS\tSCS\n");

		//---------------------------------------------------------------
		// output of real data
		//---------------------------------------------------------------

		wb_m = 256;      // start with highest WB mode

		for (k=0; k<9; k++)
			{
			printf( 
			   "%s\t%c\t%c\t%c\t%c\t%c\t%c\t%c\t%c\n",
				wb_modeName((WB_CodecMode)wb_m),     
				wb_mode_isInWBcodecSet (wb_m, data->wb_lscs) ? 'X' : '-',     
				wb_mode_isInWBcodecSet (wb_m, data->wb_lacs) ? 'X' : '-',  
				wb_mode_isInWBcodecSet (wb_m, data->wb_cacs) ? 'X' : '-',
				wb_mode_isInWBcodecSet (wb_m, data->wb_oacs) ? 'X' : '-',
				wb_mode_isInWBcodecSet (wb_m, data->wb_default_acs) ? 'X' : '-',
				wb_mode_isInWBcodecSet (wb_m, data->wb_cscs) ? 'X' : '-',
				wb_mode_isInWBcodecSet (wb_m, data->wb_dacs) ? 'X' : '-',
				wb_mode_isInWBcodecSet (wb_m, data->wb_dscs) ? 'X' : '-'
				);
			
			if (wb_m >= 2)          // just to ensure defined result in last loop
			wb_m = wb_m/2;          // moves to next lower mode
			}	
		}  
	// 5.0.0  end: output extensions for AMR-WB

	else  //5.0.0
		printf("No AMR TFO possible. \n");


	printf("\n");
	return 0;
}





//-----------------------------------------------------------------
//                  main
//-----------------------------------------------------------------

int main()
{
  Data a;
  Data* data = &a;


  while (getInputData(data))
  {
	determine_IACS_CSCS_CMACS(data);      // 5.0.0 renamed for clarity
     
	determine_OACS(data);
	 
	/*
	cout<<"After determine_OACS" << endl; 
	cout<<"lTpye: "<< data->lType <<", dType: "<< data->dType << endl; 
	cout<<" " << endl; 
	cout<<"wb_lacs: "<< data->wb_lacs <<", wb_lscs: "<< data->wb_lscs << endl; 
	cout<<"wb_dacs: "<< data->wb_dacs <<", wb_dscs: "<< data->wb_dscs << endl; 
	cout<<"wb_iacs: "<< data->wb_cacs <<", wb_cscs: "<< data->wb_cscs << endl; 
	cout<<"wb_oacs: "<< data->wb_oacs << endl; 
	cout<<" " << endl; 
	cout<<"lmacs: "<< data->lmacs <<", dmacs: "<< data->dmacs << endl; 
	cout<<"cmacs: "<< data->cmacs << endl; 
	cout<<"lOM: "<< data->lOM <<", dOM: "<< data->dOM << endl; 
	cout<<" " << endl; 
	cout<<"lacs: "<< data->lacs*1 <<", lscs: "<< data->lscs*1 << endl; 
	cout<<"dacs: "<< data->dacs*1 <<", dscs: "<< data->dscs*1 << endl; 
	cout<<"iacs: "<< data->cacs*1 <<", cscs: "<< data->cscs*1 << endl; 
	cout<<"oacs: "<< data->oacs*1 << endl; 
	// all these *1 are necessary to convert character display to a number
	cout<<" " << endl; 	// these displays were/are useful for debugging 
	*/

    printData(data);

    tfo_decision_main(data);   // also prints output

  }
  
  return 0;
}

