/*
******************************************************************************
*
*      File             : extensionsForAMRWB.h
*
*      Version          : 5.0.0  
*      Date             : 22 February 2002 
*      Author           : Clemens Suerbaum
*
*      Former Versions  : none
*
*      Description      : header file for extensions for AMR_WB
*
*
******************************************************************************
*/

#ifndef EXTENSIONSFORAMRWB_H_INCLUDED
#define EXTENSIONSFORAMRWB_H_INCLUDED

#include "oacs.h"

const unsigned short int max_wb_macs = 8;

typedef enum 
{  
	startvalue, 
	local, 
	distant
} 
LocalOrDistant;

typedef int WB_CodecSet;           

// CodecType: definition is in oacs.h (extended there to AMR-WB modes)

typedef enum                       
{
	none         = 0,
	wb_mode_0660 = 1,
	wb_mode_0885 = 2,
	wb_mode_1265 = 4,
	wb_mode_1425 = 8,
	wb_mode_1585 = 16,
	wb_mode_1825 = 32,
	wb_mode_1985 = 64,
	wb_mode_2305 = 128,
	wb_mode_2385 = 256
} 
WB_CodecMode;                      




typedef enum                       
{
  weightFactor_wb_mode_0660_GSM = 17,       // URGENT: A value MUST NOT appear more than once,
  weightFactor_wb_mode_0885_GSM = 18,       //         otherwise the algorithms later in the
  weightFactor_wb_mode_1265_GSM = 19,       //         code will fail and return nonsense ! 
  weightFactor_wb_mode_1425_GSM = 4,
  weightFactor_wb_mode_1585_GSM = 16,
  weightFactor_wb_mode_1825_GSM = 3,
  weightFactor_wb_mode_1985_GSM = 2,
  weightFactor_wb_mode_2305_GSM = 1,
  weightFactor_wb_mode_2385_GSM = 15
} 
WB_CodecWeightFactor_GSM;        // Weight factors for OHR_AMR-WB / FR_AMR-WB              

   


typedef enum                       
{
  weightFactor_wb_mode_0660_UMTS = 15,       // URGENT: A value MUST NOT appear more than once,
  weightFactor_wb_mode_0885_UMTS = 18,       //         otherwise the algorithms later in the
  weightFactor_wb_mode_1265_UMTS = 19,       //         code will fail and return nonsense ! 
  weightFactor_wb_mode_1425_UMTS = 4,
  weightFactor_wb_mode_1585_UMTS = 16,
  weightFactor_wb_mode_1825_UMTS = 3,
  weightFactor_wb_mode_1985_UMTS = 2,
  weightFactor_wb_mode_2305_UMTS = 1,
  weightFactor_wb_mode_2385_UMTS = 17
} 
WB_CodecWeightFactor_UMTS;        // Weight factors for UTRAN / 8PSK-FR              


typedef struct 
	{
	WB_CodecMode	mode;
	int				weight;
	}
Picking_Sequence_Entry;


typedef Picking_Sequence_Entry Picking_sequence[max_wb_macs];
   
// ---------------------------------------------------------------------------------- 
//                                    functions
// ---------------------------------------------------------------------------------- 

int      wb_contiguous_subset           (WB_CodecSet    subset, 
										 WB_CodecSet    superset);

int      wb_CodecSetIntersection        (WB_CodecSet    codecSet1, 
										 WB_CodecSet    codecSet2);

WB_CodecSet contiguous_WB_subset_selection (
										WB_CodecSet     wb_cs, 
									    int             lOM, 
									    int             dOM,
				                        WB_CodecSet     wb_lacs, 
									    WB_CodecSet     wb_dacs, 
									    int             cmacs
									    );

char*    wb_modeName                    (WB_CodecMode   wb_m);

char*    codecName                      (CodecType      t);

int      wb_mode_isInWBcodecSet         (int            wb_m,
										 WB_CodecSet    wb_cs);

int      lowest_mode_in_WBcodecSet      (WB_CodecSet    wb_cs);


int noOfWBmodes(WB_CodecSet wb_codecSet);

int compatible_AMR_WB_codec_types(CodecType lType, CodecType dType);

WB_CodecSet wb_contiguous_subset_selection (
	WB_CodecSet    cs, 
	WB_CodecSet    wb_lacs, 
	WB_CodecSet    wb_dacs, 
	int            cmacs);

int weight_sum_of_WBCodecSet (
							WB_CodecSet wb_cs, 
							CodecType   lType,
							CodecType   dType
							);

WB_CodecSet default_WB_codec_set (
							int cmacs, 
							CodecType   lType,
							CodecType   dType);

#endif

