/*
******************************************************************************
*
*      File             : oacs.cpp       (was: tfo_main.c)
*
*      Version          : 5.0.0  
*      Date             : 22 February 2002 
*      Author           : Clemens Suerbaum
*
*      Changes to former versions can be found by looking for the string of the
*      current version("5.0.0")
*      Changes in pure formating or for better readability of the code were not 
*      marked.
*
*      Former Version   : 2.0.0
*      Date             : 27 November 2000
*      Author           : Jochen Junkawitsch
*      Description      : source file for OACS selection rules
*
*
******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include "oacs.h"


/* ---------------------------------------------------------------------------------------
                                 private functions
   --------------------------------------------------------------------------------------- */

void      selectMode     (CodecMode codecMode, CodecSet* cscs, CodecSet* oacs);
void      notInclude     (CodecMode codecMode, CodecSet* cscs);
int       ready          (CodecSet cscs, CodecSet oacs, int cmacs);
CodecMode findHighest    (CodecSet cs);
CodecMode findLowest     (CodecSet cs);
void      selectLowest   (CodecSet* cscs, CodecSet* oacs);
void      selectHighest  (CodecSet* cscs, CodecSet* oacs);
CodecSet  calc_OACS_FR   (CodecSet cscs, int cmacs);
CodecSet  calc_OACS_HR   (CodecSet cscs, int cmacs);
void      checkDefaultFR (CodecSet* cscs, CodecSet* oacs, int cmacs);
void      checkDefaultHR (CodecSet* cscs, CodecSet* oacs, int cmacs);

/* ---------------------------------------------------------------------------------------

   --------------------------------------------------------------------------------------- */


CodecSet calc_OACS(CodecSet cscs, int cmacs, CodecType codecType)
	{
	switch (codecType)
		{
		case FR_AMR:		return calc_OACS_FR(cscs, cmacs);
		case HR_AMR:		return calc_OACS_HR(cscs, cmacs);
			
		default: fprintf(stderr, "ERROR: calc_OACS called with invalid codec type.\n"); exit(0);
		}									// 5.0.0. more detailed error text

	return (CodecType) 0;
	}


CodecSet calc_OACS_FR(CodecSet cscs, int cmacs)
{
	CodecSet oacs = 0;
    
	while (1)
	{
	  
  
	  /*
		if CMACS == 1
		select mode according to preference list {6.7, 7.4, 5.9, 5.15, 4.75, 7.95, 10.2, 12.2}
		*/
	  if (cmacs == 1)
	  {
		  selectMode(mode_670, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_740, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_590, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_515, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_475, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_795, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_102, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_122, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  fprintf(stderr, "ERROR: This point must never be reached1.\n"); exit(0);
	  }
	  
	  
	  /*
		if CMACS == 2
		*/
	  if (cmacs == 2)
	  {
		  /* If mode 10.2 is supported, do not include mode 12.2. */
		  if (cscs & mode_102) notInclude(mode_122, &cscs);
		  if (ready(cscs, oacs, cmacs)) break;
		  
		  /* Select highest mode. */
		  selectHighest(&cscs, &oacs);
		  if (ready(cscs, oacs, cmacs)) break;
		  
		  /* If mode 12.2 or mode 10.2 is selected, select mode according to
			 preference list {6.7, 7.4, 5.9, 5.15, 4.75, 7.95, 10.2, 12.2}. */
		  if (oacs & (mode_122 | mode_102))
		  {
			  selectMode(mode_670, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_740, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_590, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_515, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_475, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_795, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_102, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_122, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  }
		  
		  /* Select lowest mode. */
		  selectLowest(&cscs, &oacs);
		  if (ready(cscs, oacs, cmacs)) break;
		  
		  fprintf(stderr, "ERROR: This point must never be reached2.\n"); exit(0);
	  }
	  
	  
	  /* CMACS > 2 */
	  
	  /* if 10.2 is supported, do not include 12.2 */
	  if (cscs & mode_102) notInclude(mode_122, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* if 4.75 is supported, do not include 5.15 */
	  if (cscs & mode_475) notInclude(mode_515, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* If mode 5.15 is supported, do not include mode 5.9. */
	  if (cscs & mode_515) notInclude(mode_590, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* If mode 5.9 is supported and mode 4.75 is not supported, do not include mode 6.7. */
	  if ((cscs & mode_590) && !(cscs & mode_475)) notInclude(mode_670, &cscs); 
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* if (12.2 or 10.2) and 7.4 is supported, do not include 7.95 */
	  if ((cscs & (mode_122 | mode_102)) && (cscs & mode_740)) notInclude(mode_795, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* If mode 7.95 is supported, do not include 7.4. */
	  if (cscs & mode_795) notInclude(mode_740, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* ----- */
	  
	  /* select lowest mode */
	  selectLowest(&cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* select highest mode */
	  selectHighest(&cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* ----- */
	  
	  /* select 6.7 */
	  selectMode(mode_670, &cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* select 5.9 */
	  selectMode(mode_590, &cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  
	  /* ----- */
	  
	  fprintf(stderr, "ERROR: This point must never be reached3.\n");

  } /* end of while (1) */
  
  if (nModes(oacs) < cmacs)
	  oacs |= cscs;

  return oacs;
}


CodecSet calc_OACS_HR(CodecSet cscs, int cmacs)
{
  CodecSet oacs = 0;
    
  while (1)
  {
	  
  
	  /*
		if CMACS == 1
		select mode according to preference list {5.9, 5.15, 4.75, 6.7, 7.4, 7.95}
		*/
	  if (cmacs == 1)
	  {
		  selectMode(mode_590, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_515, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_475, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_670, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_740, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_795, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  fprintf(stderr, "ERROR: This point must never be reached4.\n"); exit(0);
	  }
	  
	  
	  /*
		if CMACS == 2
		*/
	  if (cmacs == 2)
	  {
		  /* Select highest mode. */
		  selectHighest(&cscs, &oacs);
		  if (ready(cscs, oacs, cmacs)) break;
		  
		  /* Select lowest mode. */
		  selectLowest(&cscs, &oacs);
		  if (ready(cscs, oacs, cmacs)) break;
		  
		  fprintf(stderr, "ERROR: This point must never be reached5.\n"); exit(0);
	  }
	  
	  
	  /* CMACS > 2 */
	  
	  /* if 10.2 is supported, do not include 12.2 */
	  if (cscs & mode_102) notInclude(mode_122, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* if 4.75 is supported, do not include 5.15 */
	  if (cscs & mode_475) notInclude(mode_515, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* If mode 5.15 is supported, do not include mode 5.9. */
	  if (cscs & mode_515) notInclude(mode_590, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* If mode 5.9 is supported and mode 4.75 is not supported, do not include mode 6.7. */
	  if ((cscs & mode_590) && !(cscs & mode_475)) notInclude(mode_670, &cscs); 
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* if (12.2 or 10.2) and 7.4 is supported, do not include 7.95 */
	  if ((cscs & (mode_122 | mode_102)) && (cscs & mode_740)) notInclude(mode_795, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* If mode 7.95 is supported, do not include 7.4. */
	  if (cscs & mode_795) notInclude(mode_740, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* ----- */
	  
	  /* select lowest mode */
	  selectLowest(&cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* select highest mode */
	  selectHighest(&cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* ----- */
	  
	  /* select 6.7 */
	  selectMode(mode_670, &cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* select 5.9 */
	  selectMode(mode_590, &cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  
	  /* ----- */
	  
	  fprintf(stderr, "ERROR: This point must never be reached6.\n");

  } /* end of while (1) */
  
  if (nModes(oacs) < cmacs)
	  oacs |= cscs;

  return oacs;
}


int nModes(CodecSet codecSet)
{
  int k, n=0;
  for (k=0; k<8; k++)
  { 
	  if (codecSet & 1) { n++; }
	  codecSet >>= 1;
  }
  return n;
}

void selectMode(CodecMode codecMode, CodecSet* cscs, CodecSet* oacs)
{
	if (*cscs & codecMode)           // 5.0.0 explaining comment:  if codec mode in CSCS
	{
		*cscs &= codecMode ^ 0xFF;   // 5.0.0 explaining comment: substracts codecMode from *cscs 
		*oacs |= codecMode;          // 5.0.0 explaining comment: adds codecMode to *oacs
	}
}

void notInclude(CodecMode codecMode, CodecSet* cscs)
{
  *cscs &= codecMode ^ 0xFF;
}

int ready(CodecSet cscs, CodecSet oacs, int cmacs)
{
    if (nModes(oacs) == cmacs) return 1;
    if (nModes(oacs) + nModes(cscs) <= cmacs) return 1;
    return 0;
}


CodecMode findHighest(CodecSet cs)
{
  if (cs & mode_122) return mode_122;
  if (cs & mode_102) return mode_102;
  if (cs & mode_795) return mode_795;
  if (cs & mode_740) return mode_740;
  if (cs & mode_670) return mode_670;
  if (cs & mode_590) return mode_590;
  if (cs & mode_515) return mode_515;
  if (cs & mode_475) return mode_475;
  fprintf(stderr, "ERROR: function findHighest called with empty codec set\n"); exit(0);  // 5.0.0 unique error reason
  return (CodecMode) 0;
}

CodecMode findLowest(CodecSet cs)
{
  if (cs & mode_475) return mode_475;
  if (cs & mode_515) return mode_515;
  if (cs & mode_590) return mode_590;
  if (cs & mode_670) return mode_670;
  if (cs & mode_740) return mode_740;
  if (cs & mode_795) return mode_795;
  if (cs & mode_102) return mode_102;
  if (cs & mode_122) return mode_122;
  fprintf(stderr, "ERROR: function findLowest called with empty codec set\n"); exit(0);  // 5.0.0 unique error reason
  return (CodecMode) 0;
}



void selectLowest(CodecSet* cscs, CodecSet* oacs)
{
  selectMode(findLowest(*cscs), cscs, oacs);
}

void selectHighest(CodecSet* cscs, CodecSet* oacs)
{
  selectMode(findHighest(*cscs), cscs, oacs);
}

void checkDefaultFR(CodecSet* cscs, CodecSet* oacs, int cmacs)
{
  if ((cmacs >= 4) &&
      (*cscs & mode_475) &&
      (*cscs & mode_590) &&
      (*cscs & mode_670) &&
      (*cscs & mode_102))
  {
    selectMode(mode_475, cscs, oacs);
    selectMode(mode_590, cscs, oacs);
    selectMode(mode_670, cscs, oacs);
    selectMode(mode_102, cscs, oacs);
  }
}

void checkDefaultHR(CodecSet* cscs, CodecSet* oacs, int cmacs)
{
  if ((cmacs >= 4) &&
      (*cscs & mode_475) &&
      (*cscs & mode_590) &&
      (*cscs & mode_670) &&
      (*cscs & mode_740))
  {
    selectMode(mode_475, cscs, oacs);
    selectMode(mode_590, cscs, oacs);
    selectMode(mode_670, cscs, oacs);
    selectMode(mode_740, cscs, oacs);
  }
}


// ------------------------------------------------------------------------5.0.0
//
// Function  lowModeRule
// Purpose   Checks if "high mode rule" in 28.062 is fulfilled 
// 
// ------------------------------------------------------------------------

int highModeRule(CodecSet lacs, CodecSet dacs, CodecSet oacs)
	{
	CodecMode highOACS = findHighest(oacs);
	CodecMode highLACS = findHighest(lacs);
	CodecMode highDACS = findHighest(dacs);
	CodecMode minACSMode = (highLACS < highDACS ? highLACS : highDACS);

	return (highOACS >= (minACSMode >> 1));
	}


// ------------------------------------------------------------------------5.0.0
//
// Function  lowModeRule
// Purpose   Checks if "low mode rule" in 28.062 is fulfilled 
// 
// ------------------------------------------------------------------------

int lowModeRule(CodecSet lacs, CodecSet dacs, CodecSet oacs, CodecType codecType)
	{
	if (oacs==0)  // 5.0.0 added 
		return 0;

	CodecMode max = (codecType == HR_AMR ? mode_590 : mode_740);
	
	CodecMode lowOACS = findLowest(oacs);   
	CodecMode lowLACS = findLowest(lacs);
	CodecMode lowDACS = findLowest(dacs);
	
	CodecMode lowestACSMode = (lowLACS < lowDACS ? lowLACS : lowDACS);
	
	return (lowOACS <= max || lowOACS <= lowestACSMode);
}

// ------------------------------------------------------------------------
//
// Function  preferenceList  (new in 5.0.0)
// Purpose   Prints preference of given combination
// 
// ------------------------------------------------------------------------
Preference preferenceList(CodecType lType, CodecType dType)
{
	if ( ((lType == OFR_AMR_WB) && (dType ==OFR_AMR_WB)) )
		return preference_OFR_AMR_WB__OFR_AMR_WB;

	if ( ((lType == OFR_AMR_WB) && (dType ==UMTS_AMR_WB)) ||
		 ((lType == UMTS_AMR_WB) && (dType ==OFR_AMR_WB)) )
		return preference_OFR_AMR_WB__UMTS_AMR_WB;

	if ( ((lType == UMTS_AMR_WB) && (dType ==UMTS_AMR_WB)) )
		return preference_UMTS_AMR_WB__UMTS_AMR_WB;
	
	if ( ((lType == OFR_AMR_WB) && (dType ==FR_AMR_WB)) ||
		 ((lType == FR_AMR_WB) && (dType ==OFR_AMR_WB)) )
		return preference_OFR_AMR_WB__FR_AMR_WB;
	
	if ( ((lType == FR_AMR_WB) && (dType ==UMTS_AMR_WB)) ||
		 ((lType == UMTS_AMR_WB) && (dType ==FR_AMR_WB)) )
		return preference_UMTS_AMR_WB__FR_AMR_WB;
	
	if ( ((lType == FR_AMR_WB) && (dType ==FR_AMR_WB)) )
		return preference_FR_AMR_WB__FR_AMR_WB;
	
	if ( ((lType == OFR_AMR_WB) && (dType ==OHR_AMR_WB)) ||
		 ((lType == OHR_AMR_WB) && (dType ==OFR_AMR_WB)) )
		return preference_OFR_AMR_WB__OHR_AMR_WB;
	
	if ( ((lType == UMTS_AMR_WB) && (dType ==OHR_AMR_WB)) ||
		 ((lType == OHR_AMR_WB) && (dType ==UMTS_AMR_WB)) )
		return preference_UMTS_AMR_WB__OHR_AMR_WB;
	
	if ( ((lType == FR_AMR_WB) && (dType ==OHR_AMR_WB)) ||
		 ((lType == OHR_AMR_WB) && (dType ==FR_AMR_WB)) )
		return preference_FR_AMR_WB__OHR_AMR_WB;
	
	if ( ((lType == OHR_AMR_WB) && (dType ==OHR_AMR_WB)) ||
		 ((lType == OHR_AMR_WB) && (dType ==OHR_AMR_WB)) )
		return preference_OHR_AMR_WB__OHR_AMR_WB;
	
	if ( ((lType == UMTS_AMR_2) && (dType ==UMTS_AMR_2)) )
		return preference_UMTS_AMR_2__UMTS_AMR_2;
	
	if ( ((lType == UMTS_AMR_2) && (dType ==FR_AMR)) ||
		 ((lType == FR_AMR) && (dType ==UMTS_AMR_2)) )
		return preference_UMTS_AMR_2__FR_AMR;
	
	if ( ((lType == FR_AMR) && (dType ==FR_AMR)) )
		return preference_FR_AMR__FR_AMR;
	
	if ( ((lType == UMTS_AMR_2) && (dType ==UMTS_AMR)) ||
		 ((lType == UMTS_AMR) && (dType ==UMTS_AMR_2)) )
		return preference_UMTS_AMR_2__UMTS_AMR;
	
	if ( ((lType == UMTS_AMR) && (dType ==UMTS_AMR))  )
		return preference_UMTS_AMR__UMTS_AMR;
	
	if ( ((lType == UMTS_AMR_2) && (dType ==OHR_AMR)) ||
		 ((lType == OHR_AMR) && (dType ==UMTS_AMR_2)) )
		return preference_UMTS_AMR_2__OHR_AMR;
	
	if ( ((lType == FR_AMR) && (dType ==OHR_AMR)) ||
		 ((lType == OHR_AMR) && (dType ==FR_AMR)) )
		return preference_FR_AMR__OHR_AMR;
	
	if ( ((lType == OHR_AMR) && (dType ==OHR_AMR)) )
		return preference_OHR_AMR__OHR_AMR;
	
	if ( ((lType == UMTS_AMR_2) && (dType ==HR_AMR)) ||
		 ((lType == HR_AMR) && (dType ==UMTS_AMR_2)) )
		return preference_UMTS_AMR_2__HR_AMR;
	
	if ( ((lType == FR_AMR) && (dType ==HR_AMR)) ||
		 ((lType == HR_AMR) && (dType ==FR_AMR)) )
		return preference_FR_AMR__HR_AMR;
	
	if ( ((lType == OHR_AMR) && (dType ==HR_AMR)) ||
		 ((lType == HR_AMR) && (dType ==OHR_AMR)) )
		return preference_OHR_AMR__HR_AMR;
	
	if ( ((lType == HR_AMR) && (dType ==HR_AMR)) )
		return preference_HR_AMR__HR_AMR;
	
	if ( ((lType == GSM_EFR) && (dType ==GSM_EFR)) )
		return preference_GSM_EFR__GSM_EFR;
	
	if ( ((lType == GSM_FR) && (dType ==GSM_FR)) )
		return preference_GSM_FR__GSM_FR;
	
	if ( ((lType == GSM_HR) && (dType ==GSM_HR)) )
		return preference_GSM_HR__GSM_HR;

	else
		fprintf(stderr,"ERROR: preferenceList called with unknown allowed combination\n"); 
		exit(0); 
		// return no_preference;
}