/*
******************************************************************************
*
*      File             : oacs.h
*
*      Version          : 5.0.0  
*      Date             : 22 February 2002 
*      Author           : Clemens Suerbaum
*
*      Changes to former versions can be found by looking for the string of the
*      current version("5.0.0")
*      Changes in pure formating or for better readability of the code were not 
*      marked.
*
*      Former Versions  : 2.0.0  
*      Date             : 27 February 2001
*      Author           : Jochen Junkawitsch
*
*      Description      : header file for OACS selection rules
*
*
******************************************************************************
*/

#ifndef OACS_H_INCLUDED
#define OACS_H_INCLUDED



typedef unsigned char CodecSet;


typedef enum
{
  
	dummy,                           // 5.0.0   for clearer code and easier debugging
	HR_AMR,
	FR_AMR,
	UMTS_AMR,
	UMTS_AMR_2,
	GSM_EFR,
	GSM_FR,
	GSM_HR
	,                                // 5.0.0   added  
	OHR_AMR,                         // 5.0.0   added
	FR_AMR_WB,                       // 5.0.0   added
	UMTS_AMR_WB,                     // 5.0.0   added 
	OHR_AMR_WB,                      // 5.0.0   added 
	OFR_AMR_WB                       // 5.0.0   added 
} 
CodecType;


typedef enum
{
	mode_475 = 1,
	mode_515 = 2,
	mode_590 = 4,
	mode_670 = 8,
	mode_740 = 16,
	mode_795 = 32,
	mode_102 = 64,
	mode_122 = 128
} 
CodecMode;


typedef enum         // 5.0.0 added
{
	no_preference,                           // first value  = 0
	preference_OFR_AMR_WB__OFR_AMR_WB,       // second value = 1 ...
	preference_OFR_AMR_WB__UMTS_AMR_WB,
	preference_UMTS_AMR_WB__UMTS_AMR_WB,
	preference_OFR_AMR_WB__FR_AMR_WB,	
	preference_UMTS_AMR_WB__FR_AMR_WB,
	preference_FR_AMR_WB__FR_AMR_WB,
	preference_OFR_AMR_WB__OHR_AMR_WB,	
	preference_UMTS_AMR_WB__OHR_AMR_WB,	
	preference_FR_AMR_WB__OHR_AMR_WB,	
	preference_OHR_AMR_WB__OHR_AMR_WB,	
	preference_UMTS_AMR_2__UMTS_AMR_2,
	preference_UMTS_AMR_2__FR_AMR,	
	preference_FR_AMR__FR_AMR,
	preference_UMTS_AMR_2__UMTS_AMR,	
	preference_UMTS_AMR__UMTS_AMR,
	preference_GSM_EFR__GSM_EFR,
	preference_UMTS_AMR_2__OHR_AMR,	
	preference_FR_AMR__OHR_AMR,
	preference_OHR_AMR__OHR_AMR,
	preference_UMTS_AMR_2__HR_AMR,
	preference_FR_AMR__HR_AMR,
	preference_OHR_AMR__HR_AMR,
	preference_HR_AMR__HR_AMR,
	preference_GSM_FR__GSM_FR,
	preference_GSM_HR__GSM_HR
} 
Preference;  // this implements the preference list in 28.062, section 11.6.2.


// ----------------------------------------------------------------------------------
//                                     functions
// ---------------------------------------------------------------------------------- 


CodecSet calc_OACS(CodecSet aCscs, int cmacs, CodecType codecType);
//
//   Calculate the OACS.
//   Input parameters:
//     - CodecSet aCscs     : common supported codec set (CSCS)
//     - int cmacs          : CMACS
//     - CodecType codecType: HR_AMR if HR_AMR is involved, else FR_AMR
//   Return value: OACS (type CodecSet)
//



int nModes(CodecSet codecSet);
//
//   Return the number of modes in the codecSet.
//

int highModeRule(CodecSet lacs, CodecSet dacs, CodecSet oacs);

int lowModeRule (CodecSet lacs, CodecSet dacs, CodecSet oacs, CodecType codecType);

Preference preferenceList(CodecType lType, CodecType dType);  // 5.0.0


#endif

