/*
******************************************************************************
*
*      File             : tfo_decision.c
*
*      Version          : 5.0.0  
*      Date             : 22 February 2002
*      Author           : Clemens Suerbaum
*
*      Changes to former versions can be found by looking for the string of the
*      current version("5.0.0")
*      Changes in pure formating or for better readability of the code were not 
*      marked.
*
*      Former Versions  : 2.0.0  
*      Date             : 27 February 2001
*      Author           : Jochen Junkawitsch
*
*      Description      : source file for TFO decision algorithm
*
*
******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include "tfo_decision.h"
#include "extensionsForAMRWB.h"
#include <iostream.h>                     // 5.0.0


/* ---------------------------------------------------------------------------------
                     Main TFO Decision Procedure (section 11.1)
   --------------------------------------------------------------------------------- */

int tfo_decision_main(Data* data)
{
	if (   // Two compatible AMR codec types? 
		compatible_AMRNB_codec_types(data->lType,data->dType) //5.0.0 function introduced            
		)
		{
		tfo_decision_AMR(data); // 5.0.0: no return already here because of addition
		if (acceptable(data, data->oacs))                          //5.0.0 new
			{
			cout << "Preference value: " << preferenceList(data->lType, data->dType) << endl;
			return preferenceList(data->lType, data->dType);       //5.0.0 new
			}
		else
			return 0;
		}
		
	if ( // Two compatible AMR-WB codec types?    5.0.0
		compatible_AMR_WB_codec_types(data->lType,data->dType) //5.0.0 function introduced            
		)
		{
		wb_tfo_decision_AMR(data); // 5.0.0: no return already here because of addition
		if (wb_acceptable(data, data->wb_oacs))                    //5.0.0 new
			{
			if (   
				(preferenceList(data->lType, data->dType) <= 6)
				&& 
				(preferenceList(data->lType, data->dType) > 0)    
				)
				{
				data->evaluation_wb_oacs_step1 = 
							weight_sum_of_WBCodecSet(
									data->wb_oacs,
									data->lType,
									data->dType);
				// cout << "data->evaluation_wb_oacs_step1: " << data->evaluation_wb_oacs_step1 << endl; 

				// cout << "intersect wb lacs dacs: " << wb_CodecSetIntersection(data->wb_lacs,data->wb_dacs) << endl; 
				// cout << "intersect wb oacs lacs: " << wb_CodecSetIntersection(data->wb_oacs,data->wb_lacs) << endl; 
				// cout << "intersect wb oacs dacs: " << wb_CodecSetIntersection(data->wb_oacs,data->wb_dacs) << endl; 
				// cout << "default_WB_codec_set: " << default_WB_codec_set (	noOfWBmodes(data->wb_oacs),
				//											data->lType,
				//											data->dType) << endl; 
				

				data->evaluation_wb_oacs_step2 = 
							weight_sum_of_WBCodecSet(
									wb_CodecSetIntersection(data->wb_oacs,data->wb_lacs),
									data->lType,
									data->dType)
							+ 
							weight_sum_of_WBCodecSet(
									wb_CodecSetIntersection(data->wb_oacs,data->wb_dacs),
									data->lType,
									data->dType);
				// cout << "data->evaluation_wb_oacs_step2: " << data->evaluation_wb_oacs_step2 << endl; 

				int weight_sum_wb_default_acs = weight_sum_of_WBCodecSet(
									default_WB_codec_set (	noOfWBmodes(data->wb_oacs),
															data->lType,
															data->dType),
									data->lType,
									data->dType);
				// cout << "weight_sum_wb_default_acs: " << weight_sum_wb_default_acs << endl; 

				data->evaluation_wb_oacs_step1 = data->evaluation_wb_oacs_step1 *1000/
													weight_sum_wb_default_acs;
									// *1000 for higher precision in output
				data->evaluation_wb_oacs_step2 = data->evaluation_wb_oacs_step2 *500/
													weight_sum_wb_default_acs;
									// *500 because two intersections are part of step2

				cout << "OACS evaluation step 1 (*1000): " << data->evaluation_wb_oacs_step1 << endl; 
				cout << "OACS evaluation step 2 (*1000): " << data->evaluation_wb_oacs_step2 << endl; 
				}

			cout << "Preference value: " << preferenceList(data->lType, data->dType) << endl;
			
			return preferenceList(data->lType, data->dType);       //5.0.0 new
			}
		else
			return 0;
		}

	// Same non AMR codec type? 
	if (
		data->lType == GSM_FR  && data->dType == GSM_FR  ||
		data->lType == GSM_EFR && data->dType == GSM_EFR ||
		data->lType == GSM_HR  && data->dType == GSM_HR    
		)
		{
		immediate_TFO(data); // 5.0.0: no return already here because of addition
		cout << "Preference value: " << preferenceList(data->lType, data->dType) << endl;

		return preferenceList(data->lType, data->dType);       //5.0.0 new
		}

	/* Codec Mismatch Resolution */// 5.0.0 if possible
    if (   // 5.0.0 this condition is new
		(data->lOM == 0) && (data->dOM == 0)
		)
		{
		printf("No TFO possible because of incompatible codec types and lOM=dOM=0\n");
		return 1;
		}
	else   // content of else was there before
		return codec_mismatch_resolution(data);
	}


/* ---------------------------------------------------------------------------------
                TFO Decision Algorithm for AMR Codec Types (section 11.2)
   --------------------------------------------------------------------------------- */


int tfo_decision_AMR (Data* data)
	{
	/* LACS == DACS ? */
	if (data->lacs == data->dacs)
		return immediate_TFO(data);

	/* FR - HR - Matching ? */
	if (FR_HR_matching(data))
		{
		immediate_TFO(data);
		return FR_HR_optimization(data);
		}

	/* OACS acceptable ? */
	if (!acceptable(data, data->oacs))
		return codec_mismatch_resolution(data);

	/* IACS == OACS ? */
	if (data->cacs == data->oacs)
		return immediate_TFO(data);

	/* IACS subset of OACS ? */
	if (
		contiguous_subset(data->cacs, data->oacs)
		)
		{
		immediate_TFO(data);
		return codec_mode_optimization(data);
		}

	return change_to_OACS(data);
	}



/* ---------------------------------------------------------------------------------
                          FR - HR - Matching (section 12.6)
   --------------------------------------------------------------------------------- */

int FR_HR_matching(Data* data)
{
  if (!(data->lType == FR_AMR     && data->dType == HR_AMR     ||
		data->lType == HR_AMR     && data->dType == FR_AMR     ||
		data->lType == UMTS_AMR_2 && data->dType == HR_AMR     ||
		data->lType == HR_AMR     && data->dType == UMTS_AMR_2 ||

		data->lType == OHR_AMR    && data->dType == FR_AMR     ||//5.0.0 added
		data->lType == HR_AMR     && data->dType == OHR_AMR  
		))
		         
    return 0;

  if (!data->cacs)
    return 0;

  if (!acceptable(data, data->cacs))
    return 0;

  printf("FR-HR-Matching.\n");
  return 1;
}


// ------------------------------------------------------------------------
//
// Function  determine_IACS_CSCS_CMACS
// Purpose   calculates CSCS, CMACS and IACS (directly in global data)
//           and for AMR-WB the default ACS, depending on the MACS and the codec types
// 
// ------------------------------------------------------------------------

int determine_IACS_CSCS_CMACS(Data* data)
	{
	// --------------------------------------------------------------------
	// calculate CMACS (identical for AMR and AMR-WB)
	// --------------------------------------------------------------------
	data->cmacs = data->lmacs < data->dmacs ? data->lmacs : data->dmacs;

	if (
		compatible_AMRNB_codec_types(data->lType, data->dType) //5.0.0 function used
		)	
		{
		// these parts from 2.0.0 works only for AMR-NB 
        // ----------------------------------------------------------------
		// calculate CSCS for AMR-NB
        // ----------------------------------------------------------------
		data->cscs = data->lscs & data->dscs; 

	
        // ----------------------------------------------------------------
		// calculate IACS for AMR-NB
        // ----------------------------------------------------------------
		data->cacs = contiguous_subset_selection((CodecSet)(data->lacs & data->dacs), 0,0,
						   data->lacs, data->dacs, data->cmacs);
		}

	// 5.0.0  begin: extensions for AMR-WB
	else if (
		compatible_AMR_WB_codec_types(data->lType, data->dType)
		)	
		{
        // ----------------------------------------------------------------
		// calculate CSCS for AMR-WB
        // ----------------------------------------------------------------
		data->wb_cscs = wb_CodecSetIntersection(data->wb_dscs, data->wb_lscs);
		
        // ----------------------------------------------------------------
		// calculate IACS for AMR-WB
        // ---------------------------------------------------------------- 
		WB_CodecSet intersection_of_ACSs=0;
	
		intersection_of_ACSs = wb_CodecSetIntersection(data->wb_lacs, data->wb_dacs);

		data->wb_cacs = wb_contiguous_subset_selection(
								intersection_of_ACSs,
								data->wb_lacs, 
								data->wb_dacs, 
								data->cmacs);   
		}
	// 5.0.0  end: extensions for AMR-WB
	
	return 0;
	}

// ---------------------------------------------------------------------------------
                        
// ---------------------------------------------------------------------------------

int determine_OACS(Data* data)
	{
	// no side supports ACS change 
	if (!data->lOM && !data->dOM)
		calc_OACS_case1(data);

	// both sides support ACS change 
	else if (data->lOM && data->dOM)
		{
		if (								// begin 5.0.0  
			compatible_AMR_WB_codec_types(data->lType, data->dType)
			)	
			calc_wb_OACS (data);            // end 5.0.0

		else if (      // 5.0.0 OACS calculated only for compatible AMR-NB combinations
			compatible_AMRNB_codec_types(data->lType, data->dType)
			)	
			{		
			// calculate OACS 
			data->oacs = calc_OACS( data->cscs,
									data->cmacs,
									data->lType == HR_AMR || data->dType == HR_AMR ? HR_AMR : FR_AMR);
									// 5.0.0  comment: reduces to HR / non-HR case
			}   
		}	
	
	// only one side supports ACS change 
	else
		calc_OACS_case2(data);

	if (								// begin 5.0.0  
		compatible_AMR_WB_codec_types(data->lType, data->dType)
		)	
		data->wb_default_acs = default_WB_codec_set (   noOfWBmodes(data->wb_oacs),
														data->lType,
														data->dType);

	return 0;
	}

/* ---------------------------------------------------------------------------------
                        
   --------------------------------------------------------------------------------- */

int acceptable(Data* data, CodecSet cs)
{
  CodecType codecType = data->lType == HR_AMR || data->dType == HR_AMR ? HR_AMR : FR_AMR;
  // 5.0.0 comment: this maps all other AMR-NB types to FR_AMR 
  //                needed and currently okay for lowModeRule

  return lowModeRule (data->lacs, data->dacs, cs, codecType) &&
         highModeRule(data->lacs, data->dacs, cs);;
}

/* ---------------------------------------------------------------------------------
                        
   --------------------------------------------------------------------------------- */

/* check if a is a contiguous subset of b */
int contiguous_subset(CodecSet a, CodecSet b)
{
  int mode;

  if (!a) return 0;  // empty set is never contiguous

  for (mode=mode_122; mode; mode>>=1)
  {
    if (a & mode)               // mode in a
		break;        
    b &= mode ^ 0xFF;           // substracts mode from b, if present
  }

  return a == b;
}


/* ---------------------------------------------------------------------------------
                        
   --------------------------------------------------------------------------------- */

/* section 12.2.2 Case 1: No side supports ACS change */
int calc_OACS_case1(Data* data)
	{
	if	( // Two compatible AMR codec types?           5.0.0
		compatible_AMRNB_codec_types(data->lType,data->dType) //5.0.0 function introduced            
		)
		{ 
		// 5.0.0 this block only works for AMR narrowband
		CodecSet cacs = data->lacs & data->dacs;
		if	(
			 !contiguous_subset(cacs, data->lacs) || 
			 !contiguous_subset(cacs, data->dacs)
			)
			{
			data->oacs = contiguous_subset_selection(cacs, data->lOM, data->dOM,
							 data->lacs, data->dacs, data->cmacs);
			return 0;
			}

		data->oacs = cacs;
		return 0;
		}
    
	else if	( // Two incompatible AMR codec types           5.0.0
		data->lType == UMTS_AMR   && data->dType == FR_AMR     ||
		data->lType == FR_AMR     && data->dType == UMTS_AMR   ||
		data->lType == UMTS_AMR   && data->dType == HR_AMR     ||
		data->lType == HR_AMR     && data->dType == UMTS_AMR   
		)
		{ 
		return 0;
		}

	// 5.0.0 begin extensions for AMR-WB
	else if (   
		compatible_AMR_WB_codec_types(data->lType, data->dType)
		)	
		{
		WB_CodecSet wb_cacs=0;
		wb_cacs = wb_CodecSetIntersection (data->wb_lacs, data->wb_dacs);

		if	(
			 !contiguous_subset(wb_cacs, data->wb_lacs) || 
			 !contiguous_subset(wb_cacs, data->wb_dacs)
			)
			{
			data->wb_oacs = contiguous_subset_selection(
									wb_cacs, 
									data->lOM, 
									data->dOM,
									data->wb_lacs, 
									data->wb_dacs, 
									data->cmacs);
			}

		else
			{data->wb_oacs = wb_cacs;}
		
		return 0;
		}
	               // 5.0.0 end extensions for AMR-WB

	else           // 5.0.0
		fprintf(stderr, "ERROR calc_OACS_case1 called with non-AMR codecType\n");
		return 0;
	}


/* ---------------------------------------------------------------------------------
                        
   --------------------------------------------------------------------------------- */

/* section 12.2.2 Case 2: Only one side supports ACS change */
int calc_OACS_case2(Data* data)
	{
	if	( // Two compatible AMR codec types?           5.0.0
		compatible_AMRNB_codec_types(data->lType,data->dType) //5.0.0 function introduced            
		)
		{ 
		// 5.0.0 this block only works for AMR narrowband
		CodecSet fixACS, scs, cscs;

		if (data->lOM && !data->dOM)
			{
			fixACS = data->dacs;
			scs = data->lscs;
			}
		else if (!data->lOM && data->dOM)
			{
			fixACS = data->lacs;
			scs = data->dscs;
			}
		else
			{					// 5.0.0 more detailed error text
			fprintf(stderr, "calc_OACS_case2 called with not exclusive one OM=0)\n");
			exit(0);
			}

		cscs = fixACS & scs; 
		
		if (!contiguous_subset(cscs, fixACS)) 
			{
			data->oacs = contiguous_subset_selection(cscs, data->lOM, data->dOM,
								 data->lacs, data->dacs, data->cmacs); 
			return 0;
			}

		if (nModes(cscs) > data->cmacs) 
			{

			data->oacs = contiguous_subset_selection(cscs, data->lOM, data->dOM, 
								 data->lacs, data->dacs, data->cmacs);
			return 0;
			}

		data->oacs = cscs; 

		return 0;
		}

	else if (   // AMR-WB
		compatible_AMR_WB_codec_types(data->lType, data->dType)
		)	
		               // 5.0.0 begin extensions for AMR-WB
		{
		WB_CodecSet wb_fixACS, wb_scs, wb_cscs = 0; 

		if (data->lOM && !data->dOM)
			{
			wb_fixACS = data->wb_dacs;
			wb_scs = data->wb_lscs;
			}
		else if (!data->lOM && data->dOM)
			{
			wb_fixACS = data->wb_lacs;
			wb_scs = data->wb_dscs;
			}
		else
			{								// 5.0.0 more detailed error text
			fprintf(stderr, "ERROR: calc_OACS_case2 for AMR-WB called with not exclusive one OM=0\n");  
			exit(0);
			}

		wb_cscs = wb_CodecSetIntersection (wb_fixACS, wb_scs);

		if ( 
			(noOfWBmodes(wb_cscs) > data->cmacs)
			)
			{
			data->wb_oacs = wb_contiguous_subset_selection(
						wb_cscs,					
						wb_fixACS,  // not data->wb_lacs because cont. subset from fixACS is needed 
						wb_fixACS,  // not data->wb_dacs -"- 
						data->cmacs);

			return 0;
			}

		if ( 
			!wb_contiguous_subset(wb_cscs, wb_fixACS)     
			)
			{
			data->wb_oacs = wb_contiguous_subset_selection(
									wb_cscs,             
									wb_cscs,	// not data->wb_lacs because cont. subset from fixACS is needed 
									wb_fixACS,  // not data->wb_dacs -"- 
									data->cmacs);
			return 0;
			}
		else
			{
			data->wb_oacs = wb_cscs;
			}
		}
	                   // 5.0.0 end extensions for AMR-WB

	else               // 5.0.0
		{
		fprintf(stderr, "ERROR calc_OACS_case2 called with non-AMR codecType\n");
		exit(0);
		}
	
	return 0;
	}


/* ---------------------------------------------------------------------------------
                        
   --------------------------------------------------------------------------------- */

// section 12.4 Rules for contiguous subset selection 
CodecSet contiguous_subset_selection (CodecSet cs, int lOM, int dOM,
				      CodecSet lacs, CodecSet dacs, int cmacs)
{
	int mode;
	for (mode=mode_122; mode; mode>>=1)
	{
	if (
		(lOM || contiguous_subset(cs, lacs)) &&
   		(dOM || contiguous_subset(cs, dacs)) && 
		(nModes(cs) <= cmacs)
		)
	  break;

	cs &= mode ^ 0xFF;          // subtracts mode from codec set (if present)
	}
	return cs;
}


/* ---------------------------------------------------------------------------------
                                  terminal blocks
   --------------------------------------------------------------------------------- */

int immediate_TFO(Data* data)
	{
	if (
		data->lType == FR_AMR	    ||           
		data->lType == HR_AMR	    ||
		data->lType == UMTS_AMR     || 
		data->lType == UMTS_AMR_2   
		                            ||
		data->lType == OHR_AMR
		// 5.0.0 comment: no AMR-WB codec types added here, because immediate_TFO 
		//                is only accessed for AMR-NB
		)
		printf("Immediate TFO on IACS.\n");
	else
		printf("Immediate TFO on %s.\n", codecName(data->lType));
	return 0;
	}


int codec_mismatch_resolution(Data* data)
	{
	// 5.0.0: output of inacceptable OACS added
	if (	/* Two compatible AMR codec types? */   
		compatible_AMRNB_codec_types(data->lType,data->dType) //5.0.0 function introduced            
		)
		{
		if (!acceptable(data, data->oacs))
			printf("OACS not acceptable.\n");
		}
	// 5.0.0 comment: no AMR-WB codec types added here, because codec_mismatch_resolution
	//                is only accessed for AMR-NB
		
	printf("Codec Mismatch Resolution.\n");
	return 0;
	}


int change_to_OACS(Data* data)
	{
	printf("Change ACS to OACS and establish TFO on OACS.\n");
	return 0;
	}


int codec_mode_optimization(Data* data)
	{
	printf("Codec Mode Optimization.\n");
	return 0;
	}


int FR_HR_optimization(Data* data)
	{
	printf("FR - HR - Optimization.\n");
	return 0;
	}

/* ---------------------------------------------------------------------------------
                                  help routines
   --------------------------------------------------------------------------------- */

// return a string with the name of a mode 
char* modeName(CodecMode m)
{
  switch(m)
  {
	case mode_122: return "12.20";
	case mode_102: return "10.20";
	case mode_795: return " 7.95";
	case mode_740: return " 7.40";
	case mode_670: return " 6.70";
	case mode_590: return " 5.90";
	case mode_515: return " 5.15";
	case mode_475: return " 4.75";
	default:
	  fprintf(stderr, "ERROR in modeName()\n");
	  exit(0);
  }
  return 0;
}

	  

// return a string with the name of a WB mode  //5.0.0 added
char* wb_modeName(WB_CodecMode wb_m)
{
  switch(wb_m)
  {
	case wb_mode_0660: return " 6.60";
	case wb_mode_0885: return " 8.85";
	case wb_mode_1265: return "12.65";
	case wb_mode_1425: return "14.25";
	case wb_mode_1585: return "15.85";
	case wb_mode_1825: return "18.25";
	case wb_mode_1985: return "19.85";
	case wb_mode_2305: return "23.05";
    case wb_mode_2385: return "23.85";
	default:
	  fprintf(stderr, "ERROR: conversion failed in wb_ModeName\n");
	  exit(0);
  }
  return 0;
}

	  
/* return a string with the name of a codec type */
char* codecName(CodecType t)
	{
	switch(t)
		{
		case FR_AMR:      return "FR_AMR";
		case HR_AMR:      return "HR_AMR";
		case UMTS_AMR:    return "UMTS_AMR";
		case UMTS_AMR_2:  return "UMTS_AMR_2";
		case GSM_EFR:     return "GSM_EFR";
		case GSM_FR:      return "GSM_FR";
		case GSM_HR:      return "GSM_HR";
		case OHR_AMR:     return "OHR_AMR";       // 5.0.0 added
		case FR_AMR_WB:   return "FR_AMR_WB";     // 5.0.0 added
		case UMTS_AMR_WB: return "UMTS_AMR_WB";   // 5.0.0 added
		case OFR_AMR_WB:  return "OFR_AMR_WB";    // 5.0.0 added
		case OHR_AMR_WB:  return "OHR_AMR_WB";    // 5.0.0 added
		case dummy:       return "dummy";
		default:
			fprintf(stderr, "ERROR in codecName()\n");

		exit(0);
		}
	return 0;
	}

// ------------------------------------------------------------------------
//
// Function   compatible_AMRNB_codec_types     new in 5.0.0                        
// Purpose    checks if two codec types are compatible AMR-NB codecs
//
// ------------------------------------------------------------------------
int compatible_AMRNB_codec_types(CodecType lType, CodecType dType)
	{
	if (
		( lType == UMTS_AMR_2 && dType == UMTS_AMR_2 )  ||
		( lType == UMTS_AMR   && dType == UMTS_AMR   )  ||
		( lType == FR_AMR     && dType == FR_AMR     )  ||
		( lType == HR_AMR     && dType == HR_AMR     )  ||
		( lType == FR_AMR     && dType == HR_AMR     )  ||
		( lType == HR_AMR     && dType == FR_AMR     )  ||
		( lType == UMTS_AMR_2 && dType == HR_AMR     )  ||
		( lType == HR_AMR     && dType == UMTS_AMR_2 )  ||
		( lType == UMTS_AMR_2 && dType == FR_AMR     )  ||
		( lType == FR_AMR     && dType == UMTS_AMR_2 )  ||
		( lType == UMTS_AMR_2 && dType == UMTS_AMR   )  ||
		( lType == UMTS_AMR   && dType == UMTS_AMR_2 )  ||
		( lType == UMTS_AMR   && dType == OHR_AMR    )  ||
		( lType == OHR_AMR    && dType == UMTS_AMR   )  ||
		( lType == UMTS_AMR_2 && dType == OHR_AMR    )  ||
		( lType == OHR_AMR    && dType == UMTS_AMR_2 )  ||
		( lType == FR_AMR     && dType == OHR_AMR    )  ||
		( lType == OHR_AMR    && dType == FR_AMR     )  ||
		( lType == HR_AMR     && dType == OHR_AMR    )  ||
		( lType == OHR_AMR    && dType == HR_AMR     )  ||
		( lType == OHR_AMR    && dType == OHR_AMR    )
		)
		return 1;
	else
		return 0;
	}

// ------------------------------------------------------------------------
//
// Function   calc_wb_OACS     (new in 5.0.0)                   
// Purpose    calculates the OACS if both lOM and dOM == yes
//
// ------------------------------------------------------------------------
int calc_wb_OACS (Data*   data)
	{
	if (data->wb_cscs == 0) 
		return 0;  // oacs and evalution values remains initialized on 0


	if ( noOfWBmodes(data->wb_cscs) <= data->cmacs )
		{
		data->wb_oacs = data->wb_cscs;
		return 1;
		}

	WB_CodecSet wb_oacs = 0;

	WB_CodecSet remaining_wb_CSCS = data->wb_cscs;


	int maximum_of_modes_in_WB_OACS; // minimum of: noOfModes in SCSC, max_wb_macs, wb_cmacs

	if ( noOfWBmodes(data->wb_cscs) <= max_wb_macs)
		maximum_of_modes_in_WB_OACS = noOfWBmodes(data->wb_cscs);
	else 
		maximum_of_modes_in_WB_OACS = max_wb_macs; 
	if ( maximum_of_modes_in_WB_OACS > data->cmacs ) 
		maximum_of_modes_in_WB_OACS = data->cmacs;

	int weight_0660 =0;
	int weight_0885 =0;
	int weight_1265 =0;
	int weight_1425 =0;
	int weight_1585 =0;
	int weight_1825 =0;
	int weight_1985 =0;
	int weight_2305 =0;
	int weight_2385 =0;

	int ranking_0660 =0;
	int ranking_0885 =0;
	int ranking_1265 =0;
	int ranking_1425 =0;
	int ranking_1585 =0;
	int ranking_1825 =0;
	int ranking_1985 =0;
	int ranking_2305 =0;
	int ranking_2385 =0;

	int rankingOfWBmode[8];
	int k;
	for (k=0; k<9; k++) 
		{ 
		rankingOfWBmode[k] = 0;
		}
	
	int local_weight_sum   = 0;
	int distant_weight_sum = 0;
	int oacs_weight_sum    = 0; 
	 
	double local_coefficient[max_wb_macs+1]; 
	local_coefficient[0] = 0;
	local_coefficient[1] = 0;
	local_coefficient[2] = 0;
	local_coefficient[3] = 0;
	local_coefficient[4] = 0;
	local_coefficient[5] = 0;
	local_coefficient[6] = 0;
	local_coefficient[7] = 0;
	local_coefficient[8] = 0;

	int distCoefBeatenByLocCoef[max_wb_macs];
	for (k=0; k < (max_wb_macs + 1) ; k++)  // 9 = max_wb_macs + 1
		{ 
		distCoefBeatenByLocCoef[k] = 0;
		}
	
	int pickingRankLocal[max_wb_macs];
	for (k=0; k < (max_wb_macs + 1) ; k++)  // 9 = max_wb_macs + 1
		{ 
		pickingRankLocal[k] = 0;
		}
	
	int locCoefBeatenByDistCoef[max_wb_macs];
	for (k=0; k < (max_wb_macs + 1) ; k++)  // 9 = max_wb_macs + 1
		{ 
		locCoefBeatenByDistCoef[k] = 0;
		}
	
	int pickingRankDistant[max_wb_macs];
	for (k=0; k < (max_wb_macs + 1) ; k++)  // 9 = max_wb_macs + 1
		{ 
		pickingRankDistant[k] = 0;
		}
		
	double distant_coefficient[max_wb_macs];
	for (k=0; k < (max_wb_macs + 1) ; k++)  // 9 = max_wb_macs + 1
		{ 
		distant_coefficient[k] = 0;
		}
	
	Picking_sequence local_picking_sequence; 
	for (k=0;k<9;k++)
		{
		local_picking_sequence[k].mode = none;
		local_picking_sequence[k].weight = 0;
		}
	
	Picking_sequence distant_picking_sequence; 
	for (k=0;k<9;k++)
		{
		distant_picking_sequence[k].mode = none;
		distant_picking_sequence[k].weight = 0;
		}	
	
	int ranking_locCoef1 = 0;
	
	//--------------------------------------------------------
	// for mode picking
	//--------------------------------------------------------
	int i;
	int rank;
	int noOfLocalPickedModes = 0;
	int noOfDistantPickedModes = 0;
	int noOfPickedModes = 0;
	int picked_wb_Mode = 0;
	int testMode = 1;
	int maxRanking = 0;

	//--------------------------------------------------------
	// select weight curves and prepare ranking
	//--------------------------------------------------------
	if ( // GMSK or 8PSK-HR is involved
		(data->lType==FR_AMR_WB)  || 
		(data->dType==FR_AMR_WB)  ||
		(data->lType==OHR_AMR_WB) || 
		(data->dType==OHR_AMR_WB)
		)
		{
		if (wb_mode_isInWBcodecSet(wb_mode_0660, data->wb_cscs))
			weight_0660= weightFactor_wb_mode_0660_GSM + 1;
		if (wb_mode_isInWBcodecSet(wb_mode_0885, data->wb_cscs))
			weight_0885= weightFactor_wb_mode_0885_GSM + 1;
		if (wb_mode_isInWBcodecSet(wb_mode_1265, data->wb_cscs))
			weight_1265= weightFactor_wb_mode_1265_GSM + 1;
		if (wb_mode_isInWBcodecSet(wb_mode_1425, data->wb_cscs))
			weight_1425= weightFactor_wb_mode_1425_GSM + 1;
		if (wb_mode_isInWBcodecSet(wb_mode_1585, data->wb_cscs))
			weight_1585= weightFactor_wb_mode_1585_GSM + 1;
		if (wb_mode_isInWBcodecSet(wb_mode_1825, data->wb_cscs))
			weight_1825= weightFactor_wb_mode_1825_GSM + 1;
		if (wb_mode_isInWBcodecSet(wb_mode_1985, data->wb_cscs))
			weight_1985= weightFactor_wb_mode_1985_GSM + 1;
		if (wb_mode_isInWBcodecSet(wb_mode_2305, data->wb_cscs))
			weight_2305= weightFactor_wb_mode_2305_GSM + 1;
		if (wb_mode_isInWBcodecSet(wb_mode_2385, data->wb_cscs))
			weight_2385= weightFactor_wb_mode_2385_GSM + 1;
		}					// +1 does not alter sequence, but allows 
							// exclusion of non CSCS modes later
		
	else 
		{
		if (wb_mode_isInWBcodecSet(wb_mode_0660, data->wb_cscs))
			weight_0660= weightFactor_wb_mode_0660_UMTS + 1;        
		if (wb_mode_isInWBcodecSet(wb_mode_0885, data->wb_cscs))
			weight_0885= weightFactor_wb_mode_0885_UMTS + 1;
		if (wb_mode_isInWBcodecSet(wb_mode_1265, data->wb_cscs))
			weight_1265= weightFactor_wb_mode_1265_UMTS + 1;
		if (wb_mode_isInWBcodecSet(wb_mode_1425, data->wb_cscs))
			weight_1425= weightFactor_wb_mode_1425_UMTS + 1;
		if (wb_mode_isInWBcodecSet(wb_mode_1585, data->wb_cscs))
			weight_1585= weightFactor_wb_mode_1585_UMTS + 1;
		if (wb_mode_isInWBcodecSet(wb_mode_1825, data->wb_cscs))
			weight_1825= weightFactor_wb_mode_1825_UMTS + 1;
		if (wb_mode_isInWBcodecSet(wb_mode_1985, data->wb_cscs))
			weight_1985= weightFactor_wb_mode_1985_UMTS + 1;
		if (wb_mode_isInWBcodecSet(wb_mode_2305, data->wb_cscs))
			weight_2305= weightFactor_wb_mode_2305_UMTS + 1;
		if (wb_mode_isInWBcodecSet(wb_mode_2385, data->wb_cscs))
			weight_2385= weightFactor_wb_mode_2385_UMTS + 1;
		}					// +1 does not alter sequence, but allows 
							// exclusion of non CSCS modes later
	

	//--------------------------------------------------------
	// Determine ranking of modes in CSCS
	//--------------------------------------------------------	
	ranking_0660 =	(weight_0660 > weight_0660) +
					(weight_0660 > weight_0885) +
					(weight_0660 > weight_1265) +
					(weight_0660 > weight_1425) +
					(weight_0660 > weight_1585) +
					(weight_0660 > weight_1825) +
					(weight_0660 > weight_1985) +
					(weight_0660 > weight_2305) +
					(weight_0660 > weight_2385) 
					;
	rankingOfWBmode[0] = ranking_0660;
	// cout<<"ranking_0660:" << ranking_0660 << endl; // was for debugging

	ranking_0885 =	(weight_0885 > weight_0660) +
					(weight_0885 > weight_0885) +
					(weight_0885 > weight_1265) +
					(weight_0885 > weight_1425) +
					(weight_0885 > weight_1585) +
					(weight_0885 > weight_1825) +
					(weight_0885 > weight_1985) +
					(weight_0885 > weight_2305) +
					(weight_0885 > weight_2385) 
					;
	rankingOfWBmode[1] = ranking_0885;
	// cout<<"ranking_0885:" << ranking_0885 << endl; // was for debugging

	ranking_1265 =	(weight_1265 > weight_0660) +
					(weight_1265 > weight_0885) +
					(weight_1265 > weight_1265) +
					(weight_1265 > weight_1425) +
					(weight_1265 > weight_1585) +
					(weight_1265 > weight_1825) +
					(weight_1265 > weight_1985) +
					(weight_1265 > weight_2305) +
					(weight_1265 > weight_2385) 
					;
	rankingOfWBmode[2] = ranking_1265;
	// cout<<"ranking_1265:" << ranking_1265 << endl; // was for debugging
	
	ranking_1425 =	(weight_1425 > weight_0660) +
					(weight_1425 > weight_0885) +
					(weight_1425 > weight_1265) +
					(weight_1425 > weight_1425) +
					(weight_1425 > weight_1585) +
					(weight_1425 > weight_1825) +
					(weight_1425 > weight_1985) +
					(weight_1425 > weight_2305) +
					(weight_1425 > weight_2385) 
					;
	rankingOfWBmode[3] = ranking_1425;
	// cout<<"ranking_1425:" << ranking_1425 << endl; // was for debugging
	
	ranking_1585 =	(weight_1585 > weight_0660) +
					(weight_1585 > weight_0885) +
					(weight_1585 > weight_1265) +
					(weight_1585 > weight_1425) +
					(weight_1585 > weight_1585) +
					(weight_1585 > weight_1825) +
					(weight_1585 > weight_1985) +
					(weight_1585 > weight_2305) +
					(weight_1585 > weight_2385) 
					;
	rankingOfWBmode[4] = ranking_1585;
	// cout<<"ranking_1585:" << ranking_1585 << endl; // was for debugging

	ranking_1825 =	(weight_1825 > weight_0660) +
					(weight_1825 > weight_0885) +
					(weight_1825 > weight_1265) +
					(weight_1825 > weight_1425) +
					(weight_1825 > weight_1585) +
					(weight_1825 > weight_1825) +
					(weight_1825 > weight_1985) +
					(weight_1825 > weight_2305) +
					(weight_1825 > weight_2385) 
					;
	rankingOfWBmode[5] = ranking_1825;
	// cout<<"ranking_1825:" << ranking_1825 << endl; // was for debugging

	ranking_1985 =	(weight_1985 > weight_0660) +
					(weight_1985 > weight_0885) +
					(weight_1985 > weight_1265) +
					(weight_1985 > weight_1425) +
					(weight_1985 > weight_1585) +
					(weight_1985 > weight_1825) +
					(weight_1985 > weight_1985) +
					(weight_1985 > weight_2305) +
					(weight_1985 > weight_2385) 
					;
	rankingOfWBmode[6] = ranking_1985;
	// cout<<"ranking_1985:" << ranking_1985 << endl; // was for debugging

	ranking_2305 =	(weight_2305 > weight_0660) +
					(weight_2305 > weight_0885) +
					(weight_2305 > weight_1265) +
					(weight_2305 > weight_1425) +
					(weight_2305 > weight_1585) +
					(weight_2305 > weight_1825) +
					(weight_2305 > weight_1985) +
					(weight_2305 > weight_2305) +
					(weight_2305 > weight_2385) 
					;
	rankingOfWBmode[7] = ranking_2305;
	// cout<<"ranking_2305:" << ranking_2305 << endl; // was for debugging

	ranking_2385 =	(weight_2385 > weight_0660) +
					(weight_2385 > weight_0885) +
					(weight_2385 > weight_1265) +
					(weight_2385 > weight_1425) +
					(weight_2385 > weight_1585) +
					(weight_2385 > weight_1825) +
					(weight_2385 > weight_1985) +
					(weight_2385 > weight_2305) +
					(weight_2385 > weight_2385) 
					;
	rankingOfWBmode[8] = ranking_2385;
	// cout<<"ranking_2385:" << ranking_2385 << endl; // was for debugging

	//--------------------------------------------------------
	// build weight sum
	//--------------------------------------------------------
	// local side
	if (wb_mode_isInWBcodecSet(wb_mode_0660, data->wb_lacs))
		local_weight_sum = local_weight_sum + weight_0660 - 1;  //remove 1 which was added before
	if (wb_mode_isInWBcodecSet(wb_mode_0885, data->wb_lacs))
		local_weight_sum = local_weight_sum + weight_0885 - 1;
	if (wb_mode_isInWBcodecSet(wb_mode_1265, data->wb_lacs))
		local_weight_sum = local_weight_sum + weight_1265 - 1;
	if (wb_mode_isInWBcodecSet(wb_mode_1425, data->wb_lacs))
		local_weight_sum = local_weight_sum + weight_1425 - 1;
	if (wb_mode_isInWBcodecSet(wb_mode_1585, data->wb_lacs))
		local_weight_sum = local_weight_sum + weight_1585 - 1;
	if (wb_mode_isInWBcodecSet(wb_mode_1825, data->wb_lacs))
		local_weight_sum = local_weight_sum + weight_1825 - 1;
	if (wb_mode_isInWBcodecSet(wb_mode_1985, data->wb_lacs))
		local_weight_sum = local_weight_sum + weight_1985 - 1;
	if (wb_mode_isInWBcodecSet(wb_mode_2305, data->wb_lacs))
		local_weight_sum = local_weight_sum + weight_2305 - 1;
	if (wb_mode_isInWBcodecSet(wb_mode_2385, data->wb_lacs))
		local_weight_sum = local_weight_sum + weight_2385 - 1;
	
	// distant side
	if (wb_mode_isInWBcodecSet(wb_mode_0660, data->wb_dacs))
		distant_weight_sum = distant_weight_sum + weight_0660 - 1;
	if (wb_mode_isInWBcodecSet(wb_mode_0885, data->wb_dacs))
		distant_weight_sum = distant_weight_sum + weight_0885 - 1;
	if (wb_mode_isInWBcodecSet(wb_mode_1265, data->wb_dacs))
		distant_weight_sum = distant_weight_sum + weight_1265 - 1;
	if (wb_mode_isInWBcodecSet(wb_mode_1425, data->wb_dacs))
		distant_weight_sum = distant_weight_sum + weight_1425 - 1;
	if (wb_mode_isInWBcodecSet(wb_mode_1585, data->wb_dacs))
		distant_weight_sum = distant_weight_sum + weight_1585 - 1;
	if (wb_mode_isInWBcodecSet(wb_mode_1825, data->wb_dacs))
		distant_weight_sum = distant_weight_sum + weight_1825 - 1;
	if (wb_mode_isInWBcodecSet(wb_mode_1985, data->wb_dacs))
		distant_weight_sum = distant_weight_sum + weight_1985 - 1;
	if (wb_mode_isInWBcodecSet(wb_mode_2305, data->wb_dacs))
		distant_weight_sum = distant_weight_sum + weight_2305 - 1;
	if (wb_mode_isInWBcodecSet(wb_mode_2385, data->wb_dacs))
		distant_weight_sum = distant_weight_sum + weight_2385 - 1;
		
	
	//--------------------------------------------------------
	// cover case one sum = 0 (all picking rights to other side)
	// try to avoid Handover at other side
	//--------------------------------------------------------	
	if (
		(local_weight_sum == 0) &&
		(distant_weight_sum >0)
		)
		{		
		data->wb_oacs = wb_CodecSetIntersection (data->wb_dacs, data->wb_cscs); 
				
		if ( 
			!wb_contiguous_subset(data->wb_oacs, data->wb_dacs) ||
			(noOfWBmodes(data->wb_oacs) > maximum_of_modes_in_WB_OACS)
			)
			{
			wb_oacs = wb_contiguous_subset_selection(
									 data->wb_oacs, 
				                     data->wb_dacs, 
									 data->wb_dacs, 
									 data->cmacs
									 );
			data->wb_oacs = wb_oacs;
			}

		
		if ( !wb_acceptable(data, data->wb_oacs) )  
			{ // i.e. reducing intersection did not work, Handover is needed => build from scratch
			int numberOfModes = 0;

			data->wb_oacs = 0;    // reset, because new try
			
			for (rank=8; rank>0; rank--)
				{
				if (numberOfModes < maximum_of_modes_in_WB_OACS)
					{
					if (ranking_0660==rank)
						{
						data->wb_oacs = data->wb_oacs + wb_mode_0660;
						numberOfModes = numberOfModes + 1;
						}
					else if (ranking_0885==rank) 
						{
						data->wb_oacs = data->wb_oacs + wb_mode_0885;
						numberOfModes = numberOfModes + 1;
						}
					else if (ranking_1265==rank)
						{
						data->wb_oacs = data->wb_oacs + wb_mode_1265;
						numberOfModes = numberOfModes + 1;
						}
					else if (ranking_1425==rank)
						{
						data->wb_oacs = data->wb_oacs + wb_mode_1425;
						numberOfModes = numberOfModes + 1;
						}
					else if (ranking_1585==rank)
						{
						data->wb_oacs = data->wb_oacs + wb_mode_1585;
						numberOfModes = numberOfModes + 1;
						}
					else if (ranking_1825==rank)
						{
						data->wb_oacs = data->wb_oacs + wb_mode_1825;
						numberOfModes = numberOfModes + 1;
						}
					else if (ranking_1985==rank)
						{
						data->wb_oacs = data->wb_oacs + wb_mode_1985;
						numberOfModes = numberOfModes + 1;
						}
					else if (ranking_2305==rank)
						{
						data->wb_oacs = data->wb_oacs + wb_mode_2305;
						numberOfModes = numberOfModes + 1;
						}
					else if (ranking_2385==rank)
						{
						data->wb_oacs = data->wb_oacs + wb_mode_2385;
						numberOfModes = numberOfModes + 1;
						} 
					
					} // end if number of modes < maximum_of_modes_in_WB_OACS
				} // end for rank=8..1  (this works only because MACS <= 8 )

				return 1;

			} // end if first try oacs not acceptable
			
			return 1;

		} // end if only local weight sum = 0


	if (
		(local_weight_sum > 0) &&
		(distant_weight_sum == 0)
		)
		{		
		data->wb_oacs = wb_CodecSetIntersection (data->wb_lacs, data->wb_cscs); 
				
		if ( 
			!wb_contiguous_subset(data->wb_oacs, data->wb_lacs) ||
			(noOfWBmodes(data->wb_oacs) > maximum_of_modes_in_WB_OACS)
			)
			{
			wb_oacs = wb_contiguous_subset_selection(
									 data->wb_oacs, 
				                     data->wb_lacs, 
									 data->wb_lacs, 
									 data->cmacs
									 );
			data->wb_oacs = wb_oacs;
			}

		
		if ( !wb_acceptable(data, data->wb_oacs) )  
			{ // i.e. reducing intersection did not work, Handover is needed => build from scratch
			int numberOfModes = 0;

			data->wb_oacs = 0;    // reset, because new try
			
			for (rank=8; rank>0; rank--)
				{
				if (numberOfModes < maximum_of_modes_in_WB_OACS)
					{
					if (ranking_0660==rank)
						{
						data->wb_oacs = data->wb_oacs + wb_mode_0660;
						numberOfModes = numberOfModes + 1;
						}
					else if (ranking_0885==rank) 
						{
						data->wb_oacs = data->wb_oacs + wb_mode_0885;
						numberOfModes = numberOfModes + 1;
						}
					else if (ranking_1265==rank)
						{
						data->wb_oacs = data->wb_oacs + wb_mode_1265;
						numberOfModes = numberOfModes + 1;
						}
					else if (ranking_1425==rank)
						{
						data->wb_oacs = data->wb_oacs + wb_mode_1425;
						numberOfModes = numberOfModes + 1;
						}
					else if (ranking_1585==rank)
						{
						data->wb_oacs = data->wb_oacs + wb_mode_1585;
						numberOfModes = numberOfModes + 1;
						}
					else if (ranking_1825==rank)
						{
						data->wb_oacs = data->wb_oacs + wb_mode_1825;
						numberOfModes = numberOfModes + 1;
						}
					else if (ranking_1985==rank)
						{
						data->wb_oacs = data->wb_oacs + wb_mode_1985;
						numberOfModes = numberOfModes + 1;
						}
					else if (ranking_2305==rank)
						{
						data->wb_oacs = data->wb_oacs + wb_mode_2305;
						numberOfModes = numberOfModes + 1;
						}
					else if (ranking_2385==rank)
						{
						data->wb_oacs = data->wb_oacs + wb_mode_2385;
						numberOfModes = numberOfModes + 1;
						} 
					
					} // end if number of modes < maximum_of_modes_in_WB_OACS
				} // end for rank=8..1  (this works only because MACS <= 8 )

				return 1;

			} // end if first try oacs not acceptable
			
			return 1;

		} // end if only distant weight sum = 0

	
	if (
		(local_weight_sum   == 0) &&
		(distant_weight_sum == 0)
		)
		{		
		int rank;
		int numberOfModes = 0;

		data->wb_oacs = 0;    
		
		for (rank=8; rank>0; rank--)
			{
			if (numberOfModes < maximum_of_modes_in_WB_OACS)
				{
				if (ranking_0660==rank)
					{
					data->wb_oacs = data->wb_oacs + wb_mode_0660;
					numberOfModes = numberOfModes + 1;
					}
				else if (ranking_0885==rank) 
					{
					data->wb_oacs = data->wb_oacs + wb_mode_0885;
					numberOfModes = numberOfModes + 1;
					}
				else if (ranking_1265==rank)
					{
					data->wb_oacs = data->wb_oacs + wb_mode_1265;
					numberOfModes = numberOfModes + 1;
					}
				else if (ranking_1425==rank)
					{
					data->wb_oacs = data->wb_oacs + wb_mode_1425;
					numberOfModes = numberOfModes + 1;
					}
				else if (ranking_1585==rank)
					{
					data->wb_oacs = data->wb_oacs + wb_mode_1585;
					numberOfModes = numberOfModes + 1;
					}
				else if (ranking_1825==rank)
					{
					data->wb_oacs = data->wb_oacs + wb_mode_1825;
					numberOfModes = numberOfModes + 1;
					}
				else if (ranking_1985==rank)
					{
					data->wb_oacs = data->wb_oacs + wb_mode_1985;
					numberOfModes = numberOfModes + 1;
					}
				else if (ranking_2305==rank)
					{
					data->wb_oacs = data->wb_oacs + wb_mode_2305;
					numberOfModes = numberOfModes + 1;
					}
				else if (ranking_2385==rank)
					{
					data->wb_oacs = data->wb_oacs + wb_mode_2385;
					numberOfModes = numberOfModes + 1;
					} 
				
				} // end if number of modes < maximum_of_modes_in_WB_OACS
			} // end for rank=8..1  (this works only because MACS <= 8 )

		return 1;

		} // end if both weight sums = 0


	//--------------------------------------------------------
	// 
	// GENERAL CASE: entries in both lACS and dACS
	//
	//--------------------------------------------------------
		
	//--------------------------------------------------------
	// determine coefficients for mode picking rights
	//--------------------------------------------------------
	
	for (k=1;k < max_wb_macs + 1 ; k++) 
		{
		local_coefficient[k] = (local_weight_sum * 1000 ) / ( k*2 - 1 );
		                  // multiplication to enhance precision, no influence on ranking
		// cout<<"local_coefficient["<<k<<"]: "<< local_coefficient[k] <<endl; // was for debugging
		}

	for (k=1;k < max_wb_macs + 1 ; k++) 
		{
		distant_coefficient[k] = (distant_weight_sum * 1000 ) / ( k*2 - 1 );
		// cout<<"distant_coefficient["<<k<<"]: "<< distant_coefficient[k] <<endl; // was for debugging
		}
	
	//--------------------------------------------------------
	// Determine picking order in case of coefficient equality
	//--------------------------------------------------------
    LocalOrDistant equalityPriority = startvalue;
	data->cmacs = data->lmacs < data->dmacs ? data->lmacs : data->dmacs;

	if ( // first check bigger total sum
		local_weight_sum > distant_weight_sum
		)
		equalityPriority = local;
	else if (
		local_weight_sum < distant_weight_sum
		)
		equalityPriority = distant;
	else 
		{
		for (rank=8; ((rank>0)&&(equalityPriority==startvalue)); rank--)
			{
			if (ranking_0660==rank)
				{
				if (
					(wb_mode_isInWBcodecSet(wb_mode_0660, data->wb_lacs))
					&&
					!(wb_mode_isInWBcodecSet(wb_mode_0660, data->wb_dacs))
					)
					{
					equalityPriority = local;
					printf("equalityPriority = local, wb_mode_0660\n");
					}
				else if (
					!(wb_mode_isInWBcodecSet(wb_mode_0660, data->wb_lacs))
					&&
					(wb_mode_isInWBcodecSet(wb_mode_0660, data->wb_dacs))
					)
					{
					equalityPriority = distant;
					printf("equalityPriority = distant, wb_mode_0660\n");
					}
				}
			else if (ranking_0885==rank) 
				{
				if (
					(wb_mode_isInWBcodecSet(wb_mode_0885, data->wb_lacs))
					&&
					!(wb_mode_isInWBcodecSet(wb_mode_0885, data->wb_dacs))
					)
					{
					equalityPriority = local;
					printf("equalityPriority = local, wb_mode_0885\n");
					}
				else if (
					!(wb_mode_isInWBcodecSet(wb_mode_0885, data->wb_lacs))
					&&
					(wb_mode_isInWBcodecSet(wb_mode_0885, data->wb_dacs))
					)
					{
					equalityPriority = distant;
					printf("equalityPriority = distant, wb_mode_0885\n");
					}
				}
			else if (ranking_1265==rank)
				{
				if (
					(wb_mode_isInWBcodecSet(wb_mode_1265, data->wb_lacs))
					&&
					!(wb_mode_isInWBcodecSet(wb_mode_1265, data->wb_dacs))
					)
					{
					equalityPriority = local;
					printf("equalityPriority = local, wb_mode_1265\n");
					}
				else if (
					!(wb_mode_isInWBcodecSet(wb_mode_1265, data->wb_lacs))
					&&
					(wb_mode_isInWBcodecSet(wb_mode_1265, data->wb_dacs))
					)
					{
					equalityPriority = distant;
					printf("equalityPriority = distant, wb_mode_1265\n");
					}
				}
			else if (ranking_1425==rank)
				{
				if (
					(wb_mode_isInWBcodecSet(wb_mode_1425, data->wb_lacs))
					&&
					!(wb_mode_isInWBcodecSet(wb_mode_1425, data->wb_dacs))
					)
					{
					equalityPriority = local;
					printf("equalityPriority = local, wb_mode_1425\n");
					}
				else if (
					!(wb_mode_isInWBcodecSet(wb_mode_1425, data->wb_lacs))
					&&
					(wb_mode_isInWBcodecSet(wb_mode_1425, data->wb_dacs))
					)
					{
					equalityPriority = distant;
					printf("equalityPriority = distant, wb_mode_1425\n");
					}
				}
			else if (ranking_1585==rank)
				{
				if (
					(wb_mode_isInWBcodecSet(wb_mode_1585, data->wb_lacs))
					&&
					!(wb_mode_isInWBcodecSet(wb_mode_1585, data->wb_dacs))
					)
					{
					equalityPriority = local;
					printf("equalityPriority = local, wb_mode_1585\n");
					}
				else if (
					!(wb_mode_isInWBcodecSet(wb_mode_1585, data->wb_lacs))
					&&
					(wb_mode_isInWBcodecSet(wb_mode_1585, data->wb_dacs))
					)
					{
					equalityPriority = distant;
					printf("equalityPriority = distant, wb_mode_1585\n");
					}
				}
			else if (ranking_1825==rank)
				{
				if (
					(wb_mode_isInWBcodecSet(wb_mode_1825, data->wb_lacs))
					&&
					!(wb_mode_isInWBcodecSet(wb_mode_1825, data->wb_dacs))
					)
					{
					equalityPriority = local;
					printf("equalityPriority = local, wb_mode_1825\n");
					}
				else if (
					!(wb_mode_isInWBcodecSet(wb_mode_1825, data->wb_lacs))
					&&
					(wb_mode_isInWBcodecSet(wb_mode_1825, data->wb_dacs))
					)
					{
					equalityPriority = distant;
					printf("equalityPriority = distant, wb_mode_1825\n");
					}
				}
			else if (ranking_1985==rank)
				{
				if (
					(wb_mode_isInWBcodecSet(wb_mode_1985, data->wb_lacs))
					&&
					!(wb_mode_isInWBcodecSet(wb_mode_1985, data->wb_dacs))
					)
					{
					equalityPriority = local;
					printf("equalityPriority = local, wb_mode_1985\n");
					}
				else if (
					!(wb_mode_isInWBcodecSet(wb_mode_1985, data->wb_lacs))
					&&
					(wb_mode_isInWBcodecSet(wb_mode_1985, data->wb_dacs))
					)
					{
					equalityPriority = distant;
					printf("equalityPriority = distant, wb_mode_1985\n");
					}
				}
			else if (ranking_2305==rank)
				{
				if (
					(wb_mode_isInWBcodecSet(wb_mode_2305, data->wb_lacs))
					&&
					!(wb_mode_isInWBcodecSet(wb_mode_2305, data->wb_dacs))
					)
					{
					equalityPriority = local;
					printf("equalityPriority = local, wb_mode_2305\n");
					}
				else if (
					!(wb_mode_isInWBcodecSet(wb_mode_2305, data->wb_lacs))
					&&
					(wb_mode_isInWBcodecSet(wb_mode_2305, data->wb_dacs))
					)
					{
					equalityPriority = distant;
					printf("equalityPriority = distant, wb_mode_2305\n");
					}
				}
			else if (ranking_2385==rank)
				{
				if (
					(wb_mode_isInWBcodecSet(wb_mode_2385, data->wb_lacs))
					&&
					!(wb_mode_isInWBcodecSet(wb_mode_2385, data->wb_dacs))
					)
					{
					equalityPriority = local;
					printf("equalityPriority = local, wb_mode_2385\n");
					}
				else if (
					!(wb_mode_isInWBcodecSet(wb_mode_2385, data->wb_lacs))
					&&
					(wb_mode_isInWBcodecSet(wb_mode_2385, data->wb_dacs))
					)
					{
					equalityPriority = distant;
					printf("equalityPriority = distant, wb_mode_2385\n");
					}
				}
				
			
			} // end for rank=8..1  (this works only because MACS <= 8 )

		if (equalityPriority == startvalue) // was not changed up to now
			{
			equalityPriority = local;
			} // this point can only be reached for identical ACS, 
		      // in this case the picking order is not relevant
		}

	// cout<<"equalityPriority: "<< equalityPriority <<endl; // was for debugging

	//--------------------------------------------------------
	// determine mode picking sequence
	//-------------------------------------------------------

	for (k=1; k < max_wb_macs + 1 ; k++)
		{	
		if	(local_coefficient[1] > distant_coefficient[k]) 
			distCoefBeatenByLocCoef[k] = distCoefBeatenByLocCoef[k] + 1;
		if  (local_coefficient[2] > distant_coefficient[k]) 
			distCoefBeatenByLocCoef[k] = distCoefBeatenByLocCoef[k] + 1;
		if  (local_coefficient[3] > distant_coefficient[k]) 
			distCoefBeatenByLocCoef[k] = distCoefBeatenByLocCoef[k] + 1;
		if  (local_coefficient[4] > distant_coefficient[k]) 
			distCoefBeatenByLocCoef[k] = distCoefBeatenByLocCoef[k] + 1;
		if  (local_coefficient[5] > distant_coefficient[k]) 
			distCoefBeatenByLocCoef[k] = distCoefBeatenByLocCoef[k] + 1;
		if  (local_coefficient[6] > distant_coefficient[k]) 
			distCoefBeatenByLocCoef[k] = distCoefBeatenByLocCoef[k] + 1;
		if  (local_coefficient[7] > distant_coefficient[k]) 
			distCoefBeatenByLocCoef[k] = distCoefBeatenByLocCoef[k] + 1;
		if  (local_coefficient[8] > distant_coefficient[k])
			distCoefBeatenByLocCoef[k] = distCoefBeatenByLocCoef[k] + 1;
		 
		if (
			(
			 (local_coefficient[1] == distant_coefficient[k]) ||   
			 (local_coefficient[2] == distant_coefficient[k]) ||
			 (local_coefficient[3] == distant_coefficient[k]) ||
			 (local_coefficient[4] == distant_coefficient[k]) ||
			 (local_coefficient[5] == distant_coefficient[k]) ||
			 (local_coefficient[6] == distant_coefficient[k]) ||
			 (local_coefficient[7] == distant_coefficient[k]) ||
			 (local_coefficient[8] == distant_coefficient[k])
			)
			&&
			(equalityPriority == local)
		   )
		   distCoefBeatenByLocCoef[k] = distCoefBeatenByLocCoef[k] +1;

		// cout<<"distCoefBeatenByLocCoef["<<k<<"]: "<< distCoefBeatenByLocCoef[k] <<endl; // was for debugging

		}

	for (k=1; k < max_wb_macs + 1 ; k++)
		{
		if	(local_coefficient[k] < distant_coefficient[1]) 
			locCoefBeatenByDistCoef[k] = locCoefBeatenByDistCoef[k] + 1;
		if  (local_coefficient[k] < distant_coefficient[2]) 
			locCoefBeatenByDistCoef[k] = locCoefBeatenByDistCoef[k] + 1;
		if  (local_coefficient[k] < distant_coefficient[3]) 
			locCoefBeatenByDistCoef[k] = locCoefBeatenByDistCoef[k] + 1;
		if  (local_coefficient[k] < distant_coefficient[4]) 
			locCoefBeatenByDistCoef[k] = locCoefBeatenByDistCoef[k] + 1;
		if  (local_coefficient[k] < distant_coefficient[5]) 
			locCoefBeatenByDistCoef[k] = locCoefBeatenByDistCoef[k] + 1;
		if  (local_coefficient[k] < distant_coefficient[6]) 
			locCoefBeatenByDistCoef[k] = locCoefBeatenByDistCoef[k] + 1;
		if  (local_coefficient[k] < distant_coefficient[7]) 
			locCoefBeatenByDistCoef[k] = locCoefBeatenByDistCoef[k] + 1;
		if  (local_coefficient[k] < distant_coefficient[8])
			locCoefBeatenByDistCoef[k] = locCoefBeatenByDistCoef[k] + 1;
		
		if (
			((local_coefficient[k] == distant_coefficient[1]) ||   
			 (local_coefficient[k] == distant_coefficient[2]) ||
			 (local_coefficient[k] == distant_coefficient[3]) ||
			 (local_coefficient[k] == distant_coefficient[4]) ||
			 (local_coefficient[k] == distant_coefficient[5]) ||
			 (local_coefficient[k] == distant_coefficient[6]) ||
			 (local_coefficient[k] == distant_coefficient[7]) ||
			 (local_coefficient[k] == distant_coefficient[8])
			)
			&&
			(equalityPriority == distant)
		   )
		   locCoefBeatenByDistCoef[k] = locCoefBeatenByDistCoef[k] +1;

		// cout<<"locCoefBeatenByDistCoef["<<k<<"]: "<< locCoefBeatenByDistCoef[k] <<endl; // was for debugging

		}

	for (k=1; k < max_wb_macs + 1 ; k++)
		{
		pickingRankLocal[k] = k + locCoefBeatenByDistCoef[k];
		// cout<<"pickingRankLocal["<<k<<"]: "<< pickingRankLocal[k] <<endl; 
		// was display pickingRankLocal for debugging and understanding
		}


	for (k=1; k < max_wb_macs + 1 ; k++)
		{
		pickingRankDistant[k] = k + distCoefBeatenByLocCoef[k];
		// cout<<"pickingRankDistant["<<k<<"]: "<< pickingRankDistant[k] <<endl; 
		// was display for pickingRankDistant debugging and understanding
		}

	//--------------------------------------------------------
	// exert mode picking
	//--------------------------------------------------------
	
	WB_CodecSet wb_csForLocalPicks = wb_CodecSetIntersection (data->wb_lacs, data->wb_cscs);
	// cout<<"At begin: wb_csForLocalPicks: "<< wb_csForLocalPicks <<endl; 
	
	WB_CodecSet wb_csForDistantPicks = wb_CodecSetIntersection (data->wb_dacs, data->wb_cscs);
	// cout<<"At begin: wb_csForDistantPicks: "<< wb_csForDistantPicks <<endl; 
	
	WB_CodecSet wb_csForAnyPicks = data->wb_cscs;
	// cout<<"At begin: wb_csForAnyPicks: "<< wb_csForAnyPicks <<endl; 

	int pickingRank = 0;
	// LocalOrDistant pickingSide = startvalue; // only for debugging

	for (pickingRank=1; pickingRank < maximum_of_modes_in_WB_OACS + 1 ; pickingRank++)
		{
		// cout<<"pickingRank: "<<pickingRank<<endl; //!!

		if (
			( pickingRankLocal[1] == pickingRank )   ||	
			( pickingRankLocal[2] == pickingRank )   ||	
			( pickingRankLocal[3] == pickingRank )   ||	
			( pickingRankLocal[4] == pickingRank )   ||	
			( pickingRankLocal[5] == pickingRank )   ||	
			( pickingRankLocal[6] == pickingRank )   ||	
			( pickingRankLocal[7] == pickingRank )   ||	
			( pickingRankLocal[8] == pickingRank )	
			)
			{
			// cout<<"      pickingRankLocal[k]: "<<pickingRankLocal[k]<<endl; //!!
			// pickingSide = local;
			// cout<<"      pickingSide: "<<pickingSide<<endl; //!!

			if ( wb_csForLocalPicks > 0 ) 
				{  
				testMode = 1;
				maxRanking = 0;
				
				// search for highest weigth mode
				for (i=0;i<9;i++)
					{					
					if (
						(wb_mode_isInWBcodecSet(testMode, wb_csForLocalPicks)) &&
						(rankingOfWBmode[i] > maxRanking)  
						)
						{
						picked_wb_Mode = testMode;
						maxRanking = rankingOfWBmode[i];
						}

					testMode = testMode*2; // next higher mode
					}

				}
			else if ( wb_csForDistantPicks > 0 )
				{
				testMode = 1;
				maxRanking = 0;
				
				// search for highest weigth mode
				for (i=0;i<9;i++)
					{
					if (
						(wb_mode_isInWBcodecSet(testMode, wb_csForDistantPicks)) &&
						(rankingOfWBmode[i] > maxRanking)  
						)
						{
						picked_wb_Mode = testMode;
						maxRanking = rankingOfWBmode[i];
						}

					testMode = testMode*2; // next higher mode
					}

				} // end if ( wb_csForDistantPicks > 0 )

			else if ( wb_csForAnyPicks >0 )
				{
				testMode = 1;
				maxRanking = 0;
				
				// search for highest weigth mode
				for (i=0;i<9;i++)
					{
					if (
						(wb_mode_isInWBcodecSet(testMode, wb_csForAnyPicks)) &&
						(rankingOfWBmode[i] > maxRanking)  
						)
						{
						picked_wb_Mode = testMode;
						maxRanking = rankingOfWBmode[i];
						}

					testMode = testMode*2; // next higher mode
					}

				} //  end if ( wb_csForAnyPicks > 0 )
			// search for where to pick from and which mode is done (if local may choose)
					
			
			} // end if (pickingRankLocal[k] = rank)
			

		//-----------------------------------------------------------------
		else if 
			(
			 ( pickingRankDistant[1] == pickingRank )   ||	
			 ( pickingRankDistant[2] == pickingRank )   ||	
			 ( pickingRankDistant[3] == pickingRank )   ||	
			 ( pickingRankDistant[4] == pickingRank )   ||	
			 ( pickingRankDistant[5] == pickingRank )   ||	
			 ( pickingRankDistant[6] == pickingRank )   ||	
			 ( pickingRankDistant[7] == pickingRank )   ||	
			 ( pickingRankDistant[8] == pickingRank )	
			)
			{
			// pickingSide = distant;
			// cout<<"      pickingSide: "<<pickingSide<<endl; 

			if ( wb_csForDistantPicks > 0 )
				{  
				testMode = 1;
				maxRanking = 0;
				
				// search for highest weigth mode
				for (i=0;i<9;i++)
					{
					if (
						(wb_mode_isInWBcodecSet(testMode, wb_csForDistantPicks)) &&
						(rankingOfWBmode[i] > maxRanking)  
						)
						{
						picked_wb_Mode = testMode;
						maxRanking = rankingOfWBmode[i];
						}

				 	testMode = testMode*2; // next higher mode
					}

				}  // end if ( wb_csForDistantPicks > 0 )

			else if ( wb_csForLocalPicks > 0 )
				{
				testMode = 1;
				maxRanking = 0;
				
				// search for highest weigth mode
				for (i=0;i<9;i++)
					{
					if (
						(wb_mode_isInWBcodecSet(testMode, wb_csForLocalPicks)) &&
						(rankingOfWBmode[i] > maxRanking)  
						)
						{
						picked_wb_Mode = testMode;
						maxRanking = rankingOfWBmode[i];
						}

					testMode = testMode*2; // next higher mode
					}

				} // end if ( wb_csForDistantPicks > 0 )

			else if ( wb_csForAnyPicks >0 )
				{
				testMode = 1;
				maxRanking = 0;
				
				// search for highest weigth mode
				for (i=0;i<9;i++)
					{
					if (
						(wb_mode_isInWBcodecSet(testMode, wb_csForAnyPicks)) &&
						(rankingOfWBmode[i] > maxRanking)  
						)
						{
						picked_wb_Mode = testMode;
						maxRanking = rankingOfWBmode[i];
						}

					testMode = testMode*2; // next higher mode
					}

				} // end if ( wb_csForAnyPicks >0 )
			}
			// search for where to pick from and which mode is done (if distant may choose)

		// cout << "            picked mode: " << picked_wb_Mode << endl; //!!

		data->wb_oacs = data->wb_oacs + picked_wb_Mode;
		
		noOfPickedModes = noOfPickedModes + 1;

		if (wb_mode_isInWBcodecSet(picked_wb_Mode, wb_csForLocalPicks))
			wb_csForLocalPicks   = wb_csForLocalPicks   - picked_wb_Mode;
		// cout << "After pick: wb_csForLocalPicks: " << wb_csForLocalPicks << endl; //!!

		if (wb_mode_isInWBcodecSet(picked_wb_Mode, wb_csForDistantPicks))
			wb_csForDistantPicks   = wb_csForDistantPicks   - picked_wb_Mode;
		// cout << "After pick: wb_csForDistantPicks: " << wb_csForDistantPicks << endl; //!!

		// no if because picked mode must be from wb_csForAnyPicks
		wb_csForAnyPicks  = wb_csForAnyPicks - picked_wb_Mode;
		// cout << "After pick: wb_csForAnyPicks: " << wb_csForAnyPicks << endl; //!!

		if (
			(noOfPickedModes == maximum_of_modes_in_WB_OACS) 
			||
			(wb_csForAnyPicks == 0 )
			)
			{				
			return 1;
			}
		} // for rank= 1 .. maximum_of_modes_in_WB_OACS
		
	return 0;

	} 
	


