/*
******************************************************************************
*
*      File             : tfo_decision.h
*
*      Version          : 5.0.0  
*      Date             : 22 February 2002 
*      Author           : Clemens Suerbaum
*
*      Changes to former versions can be found by looking for the string of the
*      current version("5.0.0")
*      Changes in pure formating or for better readability of the code were not 
*      marked.
*
*      Former Version   : 2.0.0
*      Date             : 15 November 2000
*      Author           : Jochen Junkawitsch
*      Description      : header file for TFO decision algorithm
*
*
******************************************************************************
*/


#ifndef TFO_DECISION_H_INCLUDED
#define TFO_DECISION_H_INCLUDED

#include "oacs.h"
#include "extensionsForAMRWB.h"   //5.0.0 added

typedef struct
{
	CodecSet lscs;
	CodecSet lacs;
	WB_CodecSet wb_lscs;    // 5.0.0 added for AMR-WB
	WB_CodecSet wb_lacs;    // 5.0.0 added for AMR-WB
	int lmacs;
	CodecType lType;
	int lOM;

	CodecSet dscs;
	CodecSet dacs;
	WB_CodecSet wb_dscs;    // 5.0.0 added for AMR-WB
	WB_CodecSet wb_dacs;    // 5.0.0 added for AMR-WB
	int dmacs;
	CodecType dType;
	int dOM;

	CodecSet oacs;
	CodecSet cscs;
	CodecSet cacs;
	WB_CodecSet wb_oacs;    // 5.0.0 added for AMR-WB
	WB_CodecSet wb_cscs;    // 5.0.0 added for AMR-WB
	WB_CodecSet wb_cacs;    // 5.0.0 added for AMR-WB
	WB_CodecSet wb_default_acs;			// 5.0.0 added for AMR-WB
	double evaluation_wb_oacs_step1;    // 5.0.0 added for AMR-WB
	double evaluation_wb_oacs_step2;    // 5.0.0 added for AMR-WB

	int cmacs;
  
} Data;


int      tfo_decision_main           (Data*    data);

int      tfo_decision_AMR            (Data*    data);

int      immediate_TFO               (Data*    data);

int      codec_mismatch_resolution   (Data*    data);

int      change_to_OACS              (Data*    data);

int      codec_mode_optimization     (Data*    data);

int      FR_HR_optimization          (Data*    data);


int      determine_OACS              (Data*    data);

int      determine_IACS_CSCS_CMACS   (Data*    data);

int      acceptable                  (Data*    data, 
									  CodecSet cs);

int      FR_HR_matching              (Data*    data);

int      contiguous_subset           (CodecSet a, 
									  CodecSet b);

int      calc_OACS_case1             (Data*    data);

int      calc_OACS_case2             (Data*    data);

CodecSet contiguous_subset_selection (CodecSet cs, 
									  int      lOM, 
									  int      dOM,
									  CodecSet lacs, 
									  CodecSet dacs, 
									  int      cmacs);

char*    modeName                    (CodecMode m);

char*    codecName                   (CodecType t);


// from here on 5.0.0
int    wb_tfo_decision_AMR			 (Data*    data);   // functions working on data-> must
                                                        // be defined in the same header file

int    wb_acceptable		         (Data*    data, 
									  WB_CodecSet cs);

int    calc_wb_OACS					 (Data*    data);

int compatible_AMRNB_codec_types(CodecType lType, CodecType dType);


#endif

